import logging
import attr
from enum import Enum
from collections import namedtuple
from .. import entities, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class ExternalStorage(str, Enum):
    """ The type of the Integration.

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - S3
         - AWS S3 drivers
       * - GCS
         - Google GCS drivers
       * - AZUREBLOB
         - Microsoft AZURE BLOB drivers
       * - AZURE_DATALAKE_GEN2
         - Microsoft AZURE GEN2 drivers
    """
    S3 = "s3"
    GCS = "gcs"
    AZUREBLOB = "azureblob"
    AZURE_DATALAKE_GEN2 = 'azureDatalakeGen2'
    POWERSCALE_S3 = 'powerscaleS3'
    MIN_IO = 's3'
    POWERSCALE_NFS = 'powerscaleNfs'


@attr.s()
class Driver(entities.BaseEntity):
    """
    Driver entity
    """
    creator = attr.ib()
    allow_external_delete = attr.ib()
    allow_external_modification = attr.ib()
    created_at = attr.ib()
    type = attr.ib()
    integration_id = attr.ib()
    integration_type = attr.ib()
    metadata = attr.ib(repr=False)
    name = attr.ib()
    id = attr.ib()
    path = attr.ib()
    bucket_name = attr.ib()
    end_point = attr.ib()
    elastic_index = attr.ib()
    elastic_index_path = attr.ib()
    directory = attr.ib()
    mount_path = attr.ib()
    # api
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['drivers'])
        return reps(
            drivers=repositories.Drivers(client_api=self._client_api),
        )

    @property
    def drivers(self):
        assert isinstance(self._repositories.drivers, repositories.Drivers)
        return self._repositories.drivers

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Driver entity object from a json

        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Driver object
        """
        inst = cls(creator=_json.get('creator', None),
                   allow_external_delete=_json.get('allowExternalDelete', None),
                   allow_external_modification=_json.get('allowExternalModification', None),
                   created_at=_json.get('createdAt', None),
                   type=_json.get('type', None),
                   integration_id=_json.get('integrationId', None),
                   integration_type=_json.get('integrationType', None),
                   metadata=_json.get('metadata', None),
                   name=_json.get('name', None),
                   id=_json.get('id', None),
                   client_api=client_api,
                   path=_json.get('path', None),
                   bucket_name=_json.get('bucketName', None),
                   end_point=_json.get('endpoint', None),
                   elastic_index=_json.get('elasticIndex', None),
                   elastic_index_path=_json.get('elasticIndexPath', None),
                   directory=_json.get('directory', None),
                   mount_path=_json.get('mountPath', None)
                   )

        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        output_dict = attr.asdict(self,
                                  filter=attr.filters.exclude(attr.fields(Driver)._client_api,
                                                              attr.fields(Driver).allow_external_delete,
                                                              attr.fields(Driver).allow_external_modification,
                                                              attr.fields(Driver).created_at,
                                                              attr.fields(Driver).integration_id,
                                                              attr.fields(Driver).integration_type,
                                                              ))
        output_dict['allowExternalDelete'] = self.allow_external_delete
        output_dict['allowExternalModification'] = self.allow_external_modification
        output_dict['createdAt'] = self.created_at
        output_dict['integrationId'] = self.integration_id
        output_dict['integrationType'] = self.integration_type

        if self.path is not None:
            output_dict['path'] = self.path

        return output_dict

    def delete(self, sure=False, really=False):
        """
        Delete a driver forever!

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param bool sure: are you sure you want to delete?
        :param bool really: really really?
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            driver.delete(sure=True, really=True)
        """
        return self.drivers.delete(driver_id=self.id,
                                   sure=sure,
                                   really=really)


@attr.s()
class AzureBlobDriver(Driver):
    container_name = attr.ib(default=None)

    def to_json(self):
        _json = super().to_json()
        _json['containerName'] = self.container_name
        return _json

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Driver entity object from a json

        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Driver object
        """
        inst = super().from_json(_json, client_api, is_fetched=True)
        inst.container_name = _json.get('containerName', None)
        return inst


@attr.s()
class GcsDriver(Driver):
    bucket = attr.ib(default=None)

    def to_json(self):
        _json = super().to_json()
        _json['bucket'] = self.bucket
        return _json

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Driver entity object from a json

        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Driver object
        """
        inst = super().from_json(_json, client_api, is_fetched=True)
        inst.bucket = _json.get('bucket', None)
        return inst


@attr.s()
class S3Driver(Driver):
    bucket_name = attr.ib(default=None)
    region = attr.ib(default=None)
    storage_class = attr.ib(default=None)

    def to_json(self):
        _json = super().to_json()
        _json['bucketName'] = self.bucket_name
        if self.region is not None:
            _json['region'] = self.region
        if self.storage_class is not None:
            _json['storageClass'] = self.storage_class
        return _json

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Driver entity object from a json

        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Driver object
        """
        inst = super().from_json(_json, client_api, is_fetched=True)
        inst.bucket_name = _json.get('bucketName', None)
        inst.region = _json.get('region', None)
        inst.storage_class = _json.get('storageClass', None)
        return inst



import traceback
from enum import Enum
from typing import Union, List
import attr
import logging

from .. import repositories, entities, exceptions

logger = logging.getLogger(name='dtlpy')


class AllocationMethod(str, Enum):
    DISTRIBUTION = 'distribution'
    PULLING = 'pulling'

class ConsensusTaskType(str, Enum):
    CONSENSUS = 'consensus'
    QUALIFICATION = 'qualification'
    HONEYPOT = 'honeypot'

class TaskPriority(int, Enum):
    LOW = 1
    MEDIUM = 2
    HIGH = 3


class ItemAction:
    def __init__(self, action, display_name=None, color='#FFFFFF', icon=None):
        if not action or not isinstance(action, str):
            raise ValueError('action should be a non-empty string')
        self.action = action
        if not display_name:
            display_name = action
        self.display_name = display_name
        self.color = color
        self.icon = icon

    @classmethod
    def from_json(cls, _json: dict):
        kwarg = {
            'action': _json.get('action')
        }

        if _json.get('displayName', False):
            kwarg['display_name'] = _json['displayName']

        if _json.get('color', False):
            kwarg['color'] = _json['color']

        if _json.get('icon', False):
            kwarg['icon'] = _json['icon']

        return cls(**kwarg)

    def to_json(self) -> dict:
        _json = {
            'action': self.action,
            'color': self.color,
            'displayName': self.display_name if self.display_name is not None else self.action
        }

        if self.icon is not None:
            _json['icon'] = self.icon

        return _json


@attr.s
class Task:
    """
    Task object
    """

    # platform
    name = attr.ib()
    status = attr.ib()
    project_id = attr.ib()
    metadata = attr.ib(repr=False)
    id = attr.ib()
    url = attr.ib(repr=False)
    task_owner = attr.ib(repr=False)
    item_status = attr.ib(repr=False)
    creator = attr.ib()
    due_date = attr.ib()
    dataset_id = attr.ib()
    spec = attr.ib()
    recipe_id = attr.ib(repr=False)
    query = attr.ib(repr=False)
    assignmentIds = attr.ib(repr=False)
    annotation_status = attr.ib(repr=False)
    progress = attr.ib()
    for_review = attr.ib()
    issues = attr.ib()
    updated_at = attr.ib()
    created_at = attr.ib()
    available_actions = attr.ib()
    total_items = attr.ib()
    priority = attr.ib()
    _description = attr.ib()

    # sdk
    _client_api = attr.ib(repr=False)
    _current_assignments = attr.ib(default=None, repr=False)
    _assignments = attr.ib(default=None, repr=False)
    _project = attr.ib(default=None, repr=False)
    _dataset = attr.ib(default=None, repr=False)
    _tasks = attr.ib(default=None, repr=False)
    _settings = attr.ib(default=None, repr=False)

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        if not isinstance(description, str):
            raise ValueError('description should be a string')
        if self._description is None:
            self._description = {}
        self._description['content'] = description

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, dataset=None):
        """
        Same as from_json but with try-except to catch if error

        :param dict _json: platform json that describe the task
        :param client_api: ApiClient object
        :param dtlpy.entities.project.Project project: project object where task will create
        :param dtlpy.entities.dataset.Dataset dataset: dataset object that refer to the task
        :return:
        """
        try:
            task = Task.from_json(
                _json=_json,
                client_api=client_api,
                project=project,
                dataset=dataset
            )
            status = True
        except Exception:
            task = traceback.format_exc()
            status = False
        return status, task

    @classmethod
    def from_json(cls, _json, client_api, project=None, dataset=None):
        """
        Return the task object form the json

        :param dict _json: platform json that describe the task
        :param client_api: ApiClient object
        :param dtlpy.entities.project.Project project: project object where task will create
        :param dtlpy.entities.dataset.Dataset dataset: dataset object that refer to the task
        :return:
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Task has been fetched from a project that is not belong to it')
                project = None

        if dataset is not None:
            if dataset.id != _json.get('datasetId', None):
                logger.warning('Task has been fetched from a dataset that is not belong to it')
                dataset = None

        actions = [ItemAction.from_json(_json=action) for action in _json.get('availableActions', list())]

        return cls(
            name=_json.get('name', None),
            status=_json.get('status', None),
            project_id=_json.get('projectId', None),
            metadata=_json.get('metadata', dict()),
            url=_json.get('url', None),
            spec=_json.get('spec', None),
            id=_json['id'],
            creator=_json.get('creator', None),
            due_date=_json.get('dueDate', 0),
            dataset_id=_json.get('datasetId', None),
            recipe_id=_json.get('recipeId', None),
            query=_json.get('query', None),
            task_owner=_json.get('taskOwner', None),
            item_status=_json.get('itemStatus', None),
            assignmentIds=_json.get('assignmentIds', list()),
            dataset=dataset,
            project=project,
            client_api=client_api,
            annotation_status=_json.get('annotationStatus', None),
            progress=_json.get('progress', None),
            for_review=_json.get('forReview', None),
            issues=_json.get('issues', None),
            updated_at=_json.get('updatedAt', None),
            created_at=_json.get('createdAt', None),
            available_actions=actions,
            total_items=_json.get('totalItems', None),
            priority=_json.get('priority', None),
            description=_json.get('description', None)
        )

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(
            self, filter=attr.filters.exclude(
                attr.fields(Task)._client_api,
                attr.fields(Task)._project,
                attr.fields(Task).project_id,
                attr.fields(Task).dataset_id,
                attr.fields(Task).recipe_id,
                attr.fields(Task).task_owner,
                attr.fields(Task).available_actions,
                attr.fields(Task).item_status,
                attr.fields(Task).due_date,
                attr.fields(Task)._tasks,
                attr.fields(Task)._dataset,
                attr.fields(Task)._current_assignments,
                attr.fields(Task)._assignments,
                attr.fields(Task).annotation_status,
                attr.fields(Task).for_review,
                attr.fields(Task).issues,
                attr.fields(Task).updated_at,
                attr.fields(Task).created_at,
                attr.fields(Task).total_items,
                attr.fields(Task)._settings,
                attr.fields(Task)._description
            )
        )
        _json['projectId'] = self.project_id
        _json['datasetId'] = self.dataset_id
        _json['recipeId'] = self.recipe_id
        _json['taskOwner'] = self.task_owner
        _json['dueDate'] = self.due_date
        _json['totalItems'] = self.total_items
        _json['forReview'] = self.for_review
        _json['description'] = self.description

        if self.available_actions is not None:
            _json['availableActions'] = [action.to_json() for action in self.available_actions]

        return _json

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/tasks/{}".format(self.project.id, self.id))

    @property
    def current_assignments(self):
        if self._current_assignments is None:
            self._current_assignments = list()
            for assignment in self.assignmentIds:
                self._current_assignments.append(self.assignments.get(assignment_id=assignment))
        return self._current_assignments

    @property
    def assignments(self):
        if self._assignments is None:
            self._assignments = repositories.Assignments(client_api=self._client_api, dataset=self._dataset,
                                                         project=self.project, task=self, project_id=self.project_id)
        assert isinstance(self._assignments, repositories.Assignments)
        return self._assignments

    @property
    def tasks(self):
        if self._tasks is None:
            self._tasks = repositories.Tasks(client_api=self._client_api, project=self.project, dataset=self.dataset)
        assert isinstance(self._tasks, repositories.Tasks)
        return self._tasks

    @property
    def settings(self):
        if self._settings is None:
            self._settings = repositories.Settings(
                client_api=self._client_api,
                project=self.project,
                dataset=self.dataset,
                task=self
            )
        assert isinstance(self._settings, repositories.Settings)
        return self._settings

    @property
    def project(self):
        if self._project is None:
            self.get_project()
            if self._project is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Missing entity "project". need to "get_project()" ')
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def dataset(self):
        if self._dataset is None:
            self.get_dataset()
            if self._dataset is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Missing entity "dataset". need to "get_dataset()" ')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    def get_project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id)

    def get_dataset(self):
        if self._dataset is None:
            self._dataset = repositories.Datasets(client_api=self._client_api, project=self._project).get(
                dataset_id=self.dataset_id)

    def open_in_web(self):
        """
        Open the task in web platform

        :return:
        """
        self._client_api._open_in_web(url=self.platform_url)

    def delete(self, wait=True):
        """
        Delete task from platform

        :param bool wait: wait until delete task finish
        :return: True
        :rtype: bool
        """
        return self.tasks.delete(task_id=self.id, wait=wait)

    def update(self, system_metadata=False):
        """
        Update an Annotation Task

        :param bool system_metadata: DEPRECATED
        """
        return self.tasks.update(task=self, system_metadata=system_metadata)

    def create_qa_task(self,
                       due_date,
                       assignee_ids,
                       filters=None,
                       items=None,
                       query=None,
                       workload=None,
                       metadata=None,
                       available_actions=None,
                       wait=True,
                       batch_size=None,
                       max_batch_workload=None,
                       allowed_assignees=None,
                       priority=TaskPriority.MEDIUM
                       ):
        """
        Create a new QA Task

        :param float due_date: date by which the QA task should be finished; for example, due_date=datetime.datetime(day=1, month=1, year=2029).timestamp()
        :param list assignee_ids: list the QA task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param entities.Filters filters: dl.Filters entity to filter items for the task
        :param List[entities.Item] items: list of items (item Id or objects) to insert to the task
        :param dict DQL query: filter items for the task
        :param List[WorkloadUnit] workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param dict metadata: metadata for the task
        :param list available_actions: list of available actions (statuses) that will be available for the task items; The default statuses are: "approved" and "discard"
        :param bool wait: wait until create task finish
        :param int batch_size: Pulling batch size (items), use with pulling allocation method. Restrictions - Min 3, max 100
        :param int max_batch_workload: Max items in assignment, use with pulling allocation method. Restrictions - Min batchSize + 2, max batchSize * 2
        :param list allowed_assignees: list the task assignees (contributors) that should be working on the task. Provide a list of users' emails
        :param entities.TaskPriority priority: priority of the task options in entities.TaskPriority
        :return: task object
        :rtype: dtlpy.entities.task.Task

        **Example**:

        .. code-block:: python

            task = task.create_qa_task(due_date = datetime.datetime(day= 1, month= 1, year= 2029).timestamp(),
                                assignee_ids =[ 'annotator1@dataloop.ai', 'annotator2@dataloop.ai'])
        """
        return self.tasks.create_qa_task(task=self,
                                         due_date=due_date,
                                         assignee_ids=assignee_ids,
                                         filters=filters,
                                         items=items,
                                         query=query,
                                         workload=workload,
                                         metadata=metadata,
                                         available_actions=available_actions,
                                         wait=wait,
                                         batch_size=batch_size,
                                         max_batch_workload=max_batch_workload,
                                         allowed_assignees=allowed_assignees,
                                         priority=priority
                                         )

    def create_assignment(self, assignment_name, assignee_id, items=None, filters=None):
        """
        Create a new assignment

        :param str assignment_name: assignment name
        :param str assignee_id: the assignment assignees (contributors) that should be working on the task. Provide a user email
        :param List[entities.Item] items: list of items (item Id or objects) to insert to the task
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment assignment

        **Example**:

        .. code-block:: python

            assignment = task.create_assignment(assignee_id='annotator1@dataloop.ai')
        """
        assignment = self.assignments.create(assignee_id=assignee_id,
                                             filters=filters,
                                             items=items)

        assignment.metadata['system']['taskId'] = self.id
        assignment.update(system_metadata=True)
        self.assignmentIds.append(assignment.id)
        self.update()
        self.add_items(filters=filters, items=items)
        return assignment

    def add_items(self, filters=None, items=None, assignee_ids=None, workload=None, limit=None, wait=True, query=None):
        """
        Add items to Task

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param list items: list of items (item Ids or objects) to add to the task
        :param list assignee_ids: list to assignee who works in the task
        :param list workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param int limit: the limit items that task can include
        :param bool wait: wait until add items will to finish
        :param dict query: query to filter the items for the task

        :return: task entity
        :rtype: dtlpy.entities.task.Task
        """
        return self.tasks.add_items(task=self,
                                    filters=filters,
                                    items=items,
                                    assignee_ids=assignee_ids,
                                    workload=workload,
                                    limit=limit,
                                    wait=wait,
                                    query=query)

    def remove_items(self,
                     filters: entities.Filters = None,
                     query=None,
                     items=None,
                     wait=True):
        """
        remove items from Task.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned to be *owner* of the annotation task.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param dict query: query to filter the items use it
        :param list items: list of items to add to the task
        :param bool wait: wait until remove items finish

        :return: True if success and an error if failed
        :rtype: bool
        """
        return self.tasks.remove_items(task=self,
                                       query=query,
                                       filters=filters,
                                       items=items,
                                       wait=wait)

    def get_items(self, filters=None, get_consensus_items: bool = False):
        """
        Get the task items

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: list of the items or PagedEntity output of items
        :rtype: list or dtlpy.entities.paged_entities.PagedEntities
        """
        return self.tasks.get_items(task_id=self.id, dataset=self.dataset, filters=filters, get_consensus_items=get_consensus_items)

    def set_status(self, status: str, operation: str, item_ids: List[str]):
        """
        Update item status within task

        :param str status: string the describes the status
        :param str operation: the status action need 'create' or 'delete'
        :param list item_ids: List[str] id items ids

        :return: True if success
        :rtype: bool
        """
        return self.tasks.set_status(status=status, operation=operation, item_ids=item_ids, task_id=self.id)
    



from copy import deepcopy
import traceback
import logging
import json
import os
from PIL import Image
from io import BytesIO
import base64

from .. import entities, exceptions, miscellaneous, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Annotations:
    """
    Annotations Repository

    The Annotation class allows you to manage the annotations of data items. For information on annotations explore our
    documentation at:
    `Classification SDK <https://developers.dataloop.ai/tutorials/annotations_image/classification_point_and_pose/chapter/>`_,
    `Annotation Labels and Attributes <https://developers.dataloop.ai/tutorials/data_management/upload_and_manage_annotations/chapter/#set-attributes-on-annotations>`_,
    `Show Video with Annotations <https://developers.dataloop.ai/tutorials/annotations_video/video_annotations/chapter/>`_.
    """

    def __init__(self, client_api: ApiClient, item=None, dataset=None, dataset_id=None):
        self._client_api = client_api
        self._item = item
        self._dataset = dataset
        self._upload_batch_size = 100
        if dataset_id is None:
            if dataset is not None:
                dataset_id = dataset.id
            elif item is not None:
                dataset_id = item.dataset_id
        self._dataset_id = dataset_id

    ############
    # entities #
    ############
    @property
    def dataset(self):
        if self._dataset is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "dataset". need to set a Dataset entity or use dataset.annotations repository')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    @property
    def item(self):
        if self._item is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "item". need to set an Item entity or use item.annotations repository')
        assert isinstance(self._item, entities.Item)
        return self._item

    @item.setter
    def item(self, item: entities.Item):
        if not isinstance(item, entities.Item):
            raise ValueError('Must input a valid Item entity')
        self._item = item

    ###########
    # methods #
    ###########
    @_api_reference.add(path='/annotations/{annotationId}', method='get')
    def get(self, annotation_id: str) -> entities.Annotation:
        """
        Get a single annotation.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
        *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param str annotation_id: The id of the annotation
        :return: Annotation object or None
        :return: Annotation object or None
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = item.annotations.get(annotation_id='annotation_id')
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/annotations/{}'.format(annotation_id))
        if success:
            annotation = entities.Annotation.from_json(_json=response.json(),
                                                       annotations=self,
                                                       dataset=self._dataset,
                                                       client_api=self._client_api,
                                                       item=self._item)
        else:
            raise exceptions.PlatformException(response)
        return annotation

    def _build_entities_from_response(self, response_items):
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_json, _json in enumerate(response_items):
            jobs[i_json] = pool.submit(entities.Annotation._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': _json,
                                          'item': self._item,
                                          'dataset': self._dataset,
                                          'annotations': self})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        """
        Get a dataset's item list. This is a browsing endpoint. For any given path, item count will be returned.
        The user is then expected to perform another request for every folder to actually get its item list.

        :param dtlpy.entities.filters.Filters filters: Filter entity or a dictionary containing filters parameters

        :return: json response
        :rtype: 
        """
        # prepare request
        success, response = self._client_api.gen_request(req_type="POST",
                                                         path="/datasets/{}/query".format(self._dataset_id),
                                                         json_req=filters.prepare(),
                                                         headers={'user_query': filters._user_query}
                                                         )
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/datasets/{id}/query', method='post')
    def list(self, filters: entities.Filters = None, page_offset: int = None, page_size: int = None):
        """
        List Annotations of a specific item. You must get the item first and then list the annotations with the desired filters.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
        *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param int page_offset: starting page
        :param int page_size: size of page
        :return: Pages object
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            annotations = item.annotations.list(filters=dl.Filters(
                                         resource=dl.FiltersResource.ANNOTATION,
                                         field='type',
                                         values='box'),
                      page_size=100,
                      page_offset=0)
        """
        if self._dataset_id is not None:
            if filters is None:
                filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION)
                filters._user_query = 'false'

            if not filters.resource == entities.FiltersResource.ANNOTATION:
                raise exceptions.PlatformException(error='400',
                                                   message='Filters resource must to be FiltersResource.ANNOTATION')

            if self._item is not None and not filters.has_field('itemId'):
                filters = deepcopy(filters)
                filters.page_size = 1000
                filters.add(field='itemId', values=self.item.id, method=entities.FiltersMethod.AND)

            # assert type filters
            if not isinstance(filters, entities.Filters):
                raise exceptions.PlatformException('400', 'Unknown filters type')

            # page size
            if page_size is None:
                # take from default
                page_size = filters.page_size
            else:
                filters.page_size = page_size

            # page offset
            if page_offset is None:
                # take from default
                page_offset = filters.page
            else:
                filters.page = page_offset

            paged = entities.PagedEntities(items_repository=self,
                                           filters=filters,
                                           page_offset=page_offset,
                                           page_size=page_size,
                                           client_api=self._client_api)
            paged.get_page()

            if self._item is not None:
                if paged.total_pages_count > 1:
                    annotations = list()
                    for page in paged:
                        annotations += page
                else:
                    annotations = paged.items
                return entities.AnnotationCollection(annotations=annotations, item=self._item)
            else:
                return paged
        else:
            raise exceptions.PlatformException('400',
                                               'Please use item.annotations.list() or dataset.annotations.list() '
                                               'to perform this action.')

    def show(self,
             image=None,
             thickness: int = 1,
             with_text: bool = False,
             height: float = None,
             width: float = None,
             annotation_format: entities.ViewAnnotationOptions = entities.ViewAnnotationOptions.MASK,
             alpha: float = 1):
        """
        Show annotations. To use this method, you must get the item first and then show the annotations with
        the desired filters. The method returns an array showing all the annotations.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
        *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param ndarray image: empty or image to draw on
        :param int thickness: optional - line thickness, default=1
        :param bool with_text: add label to annotation
        :param float height: item height
        :param float width: item width
        :param str annotation_format: the format that want to show ,options: list(dl.ViewAnnotationOptions)
        :param float alpha: opacity value [0 1], default 1
        :return: ndarray of the annotations
        :rtype: ndarray

        **Example**:

        .. code-block:: python

            image = item.annotations.show(image='nd array',
                      thickness=1,
                      with_text=False,
                      height=100,
                      width=100,
                      annotation_format=dl.ViewAnnotationOptions.MASK,
                      alpha=1)
        """
        # get item's annotations
        annotations = self.list()

        return annotations.show(image=image,
                                width=width,
                                height=height,
                                thickness=thickness,
                                alpha=alpha,
                                with_text=with_text,
                                annotation_format=annotation_format)

    def download(self,
                 filepath: str,
                 annotation_format: entities.ViewAnnotationOptions = entities.ViewAnnotationOptions.JSON,
                 img_filepath: str = None,
                 height: float = None,
                 width: float = None,
                 thickness: int = 1,
                 with_text: bool = False,
                 alpha: float = 1):
        """
        Save annotation to file.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
        *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param str filepath: Target download directory
        :param str annotation_format: the format that want to download ,options: list(dl.ViewAnnotationOptions)
        :param str img_filepath: img file path - needed for img_mask
        :param float height: optional - image height
        :param float width: optional - image width
        :param int thickness: optional - line thickness, default=1
        :param bool with_text: optional - draw annotation with text, default = False
        :param float alpha: opacity value [0 1], default 1
        :return: file path to where save the annotations
        :rtype: str

        **Example**:

        .. code-block:: python

            file_path = item.annotations.download(
                          filepath='file_path',
                          annotation_format=dl.ViewAnnotationOptions.MASK,
                          img_filepath='img_filepath',
                          height=100,
                          width=100,
                          thickness=1,
                          with_text=False,
                          alpha=1)
        """
        # get item's annotations
        annotations = self.list()
        if 'text' in self.item.metadata.get('system').get('mimetype', '') or 'json' in self.item.metadata.get('system').get('mimetype', ''):
            annotation_format = entities.ViewAnnotationOptions.JSON
        elif 'audio' not in self.item.metadata.get('system').get('mimetype', ''):
            # height/weight
            if height is None:
                if self.item.height is None:
                    raise exceptions.PlatformException('400', 'Height must be provided')
                height = self.item.height
            if width is None:
                if self.item.width is None:
                    raise exceptions.PlatformException('400', 'Width must be provided')
                width = self.item.width

        return annotations.download(filepath=filepath,
                                    img_filepath=img_filepath,
                                    width=width,
                                    height=height,
                                    thickness=thickness,
                                    with_text=with_text,
                                    annotation_format=annotation_format,
                                    alpha=alpha)

    def _delete_single_annotation(self, w_annotation_id):
        try:
            payload = {'username': self._client_api.info()['user_email']}
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path='/annotations/{}'.format(w_annotation_id),
                                                             json_req=payload)

            if not success:
                raise exceptions.PlatformException(response)
            return success
        except Exception:
            logger.exception('Failed to delete annotation')
            raise

    @_api_reference.add(path='/annotations/{annotationId}', method='delete')
    def delete(self, annotation: entities.Annotation = None,
               annotation_id: str = None,
               filters: entities.Filters = None) -> bool:
        """
        Remove an annotation from item.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
        *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param dtlpy.entities.annotation.Annotation annotation: Annotation object
        :param str annotation_id: The id of the annotation
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: True/False
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_deleted = item.annotations.delete(annotation_id='annotation_id')
        """
        if annotation is not None:
            if isinstance(annotation, entities.Annotation):
                annotation_id = annotation.id
            elif isinstance(annotation, str) and annotation.lower() == 'all':
                if self._item is None:
                    raise exceptions.PlatformException(error='400',
                                                       message='To use "all" option repository must have an item')
                filters = entities.Filters(
                    resource=entities.FiltersResource.ANNOTATION,
                    field='itemId',
                    values=self._item.id,
                    method=entities.FiltersMethod.AND
                )
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='Unknown annotation type')

        if annotation_id is not None:
            if not isinstance(annotation_id, list):
                return self._delete_single_annotation(w_annotation_id=annotation_id)
            filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION,
                                       field='annotationId',
                                       values=annotation_id,
                                       operator=entities.FiltersOperations.IN)
            filters.pop(field="type")

        if filters is None:
            raise exceptions.PlatformException(error='400',
                                               message='Must input filter, annotation id or annotation entity')

        if not filters.resource == entities.FiltersResource.ANNOTATION:
            raise exceptions.PlatformException(error='400',
                                               message='Filters resource must to be FiltersResource.ANNOTATION')

        if self._item is not None and not filters.has_field('itemId'):
            filters = deepcopy(filters)
            filters.add(field='itemId', values=self._item.id, method=entities.FiltersMethod.AND)

        if self._dataset is not None:
            items_repo = self._dataset.items
        elif self._item is not None:
            items_repo = self._item.dataset.items
        else:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "dataset". need to set a Dataset entity or use dataset.annotations repository')

        return items_repo.delete(filters=filters)

    @staticmethod
    def _update_snapshots(origin, modified):
        """
        Update the snapshots if a change occurred and return a list of the new snapshots with the flag to update.
        """
        update = False
        origin_snapshots = list()
        origin_metadata = origin['metadata'].get('system', None)
        if origin_metadata:
            origin_snapshots = origin_metadata.get('snapshots_', None)
            if origin_snapshots is not None:
                modified_snapshots = modified['metadata'].get('system', dict()).get('snapshots_', None)
                # if the number of the snapshots change
                if len(origin_snapshots) != len(modified_snapshots):
                    origin_snapshots = modified_snapshots
                    update = True

                i = 0
                # if some snapshot change
                while i < len(origin_snapshots) and not update:
                    if origin_snapshots[i] != modified_snapshots[i]:
                        origin_snapshots = modified_snapshots
                        update = True
                        break
                    i += 1

        return update, origin_snapshots

    def _update_single_annotation(self, w_annotation, system_metadata):
        try:
            if isinstance(w_annotation, entities.Annotation):
                if w_annotation.id is None:
                    raise exceptions.PlatformException(
                        '400',
                        'Cannot update annotation because it was not fetched'
                        ' from platform and therefore does not have an id'
                    )
                annotation_id = w_annotation.id
            else:
                raise exceptions.PlatformException('400',
                                                   'unknown annotations type: {}'.format(type(w_annotation)))

            origin = w_annotation._platform_dict
            modified = w_annotation.to_json()
            # check snapshots
            update, updated_snapshots = self._update_snapshots(origin=origin,
                                                               modified=modified)

            # pop the snapshots to make the diff work with out them
            origin.get('metadata', dict()).get('system', dict()).pop('snapshots_', None)
            modified.get('metadata', dict()).get('system', dict()).pop('snapshots_', None)

            # check diffs in the json
            json_req = miscellaneous.DictDiffer.diff(origin=origin,
                                                     modified=modified)

            # add the new snapshots if exist
            if updated_snapshots and update:
                if 'metadata' not in json_req:
                    json_req['metadata'] = dict()
                if 'system' not in json_req['metadata']:
                    json_req['metadata']['system'] = dict()
                json_req['metadata']['system']['snapshots_'] = updated_snapshots

            # no changes happen
            if not json_req and not updated_snapshots:
                status = True
                result = w_annotation
            else:
                suc, response = self._update_annotation_req(annotation_json=json_req,
                                                            system_metadata=system_metadata,
                                                            annotation_id=annotation_id)
                if suc:
                    result = entities.Annotation.from_json(_json=response.json(),
                                                           annotations=self,
                                                           dataset=self._dataset,
                                                           item=self._item)
                    w_annotation._platform_dict = result._platform_dict
                else:
                    raise exceptions.PlatformException(response)
                status = True
        except Exception:
            status = False
            result = traceback.format_exc()
        return status, result

    def _update_annotation_req(self, annotation_json, system_metadata, annotation_id):
        url_path = '/annotations/{}'.format(annotation_id)
        if system_metadata:
            url_path += '?system=true'
        suc, response = self._client_api.gen_request(req_type='put',
                                                     path=url_path,
                                                     json_req=annotation_json)
        return suc, response

    @_api_reference.add(path='/annotations/{annotationId}', method='put')
    def update(self, annotations, system_metadata=False):
        """
        Update an existing annotation. For example, you may change the annotation's label and then use the update method. 

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
         *developer* or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :param dtlpy.entities.annotation.Annotation annotations: Annotation object
        :param bool system_metadata: bool - True, if you want to change metadata system

        :return: True if successful or error if unsuccessful
        :rtype: bool

         **Example**:

        .. code-block:: python

            annotations = item.annotations.update(annotation='annotation')
        """
        pool = self._client_api.thread_pools(pool_name='annotation.update')
        if not isinstance(annotations, list):
            annotations = [annotations]
        jobs = [None for _ in range(len(annotations))]
        for i_ann, ann in enumerate(annotations):
            jobs[i_ann] = pool.submit(self._update_single_annotation,
                                      **{'w_annotation': ann,
                                         'system_metadata': system_metadata})

        # get all results
        results = [j.result() for j in jobs]
        out_annotations = [r[1] for r in results if r[0] is True]
        out_errors = [r[1] for r in results if r[0] is False]
        if len(out_errors) == 0:
            logger.debug('Annotation/s updated successfully. {}/{}'.format(len(out_annotations), len(results)))
        else:
            logger.error(out_errors)
            logger.error('Annotation/s updated with {} errors'.format(len(out_errors)))
        return out_annotations

    @staticmethod
    def _annotation_encoding(annotation):
        metadata = annotation.get('metadata', dict())
        system = metadata.get('system', dict())
        snapshots = system.get('snapshots_', list())
        last_frame = {
            'label': annotation.get('label', None),
            'attributes': annotation.get('attributes', None),
            'type': annotation.get('type', None),
            'data': annotation.get('coordinates', None),
            'fixed': snapshots[0].get('fixed', None) if (isinstance(snapshots, list) and len(snapshots) > 0) else None,
            'objectVisible': snapshots[0].get('objectVisible', None) if (
                    isinstance(snapshots, list) and len(snapshots) > 0) else None,
        }

        offset = 0
        for idx, frame in enumerate(deepcopy(snapshots)):
            frame.pop("frame", None)
            if frame == last_frame and not frame['fixed']:
                del snapshots[idx - offset]
                offset += 1
            else:
                last_frame = frame
        return annotation

    def _create_batches_for_upload(self, annotations, merge=False):
        """
        receives a list of annotations and split them into batches to optimize the upload

        :param annotations: list of all annotations
        :param merge: bool - merge the new binary annotations with the existing annotations
        :return: batch_annotations: list of list of annotation. each batch with size self._upload_batch_size
        """
        annotation_batches = list()
        single_batch = list()
        for annotation in annotations:
            if isinstance(annotation, str):
                annotation = json.loads(annotation)
            elif isinstance(annotation, entities.Annotation):
                if annotation._item is None and self._item is not None:
                    # if annotation is without item - set one (affects the binary annotation color)
                    annotation._item = self._item
                annotation = annotation.to_json()
            elif isinstance(annotation, dict):
                pass
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='unknown annotations type: {}'.format(type(annotation)))
            annotation = self._annotation_encoding(annotation)
            single_batch.append(annotation)
            if len(single_batch) >= self._upload_batch_size:
                annotation_batches.append(single_batch)
                single_batch = list()
        if len(single_batch) > 0:
            annotation_batches.append(single_batch)
        if merge and self.item:
            annotation_batches = self._merge_new_annotations(annotation_batches)
            annotation_batches = self._merge_to_exits_annotations(annotation_batches)
        return annotation_batches

    def _merge_binary_annotations(self, data_url1, data_url2, item_width, item_height):
        # Decode base64 data
        img_data1 = base64.b64decode(data_url1.split(",")[1])
        img_data2 = base64.b64decode(data_url2.split(",")[1])

        # Convert binary data to images
        img1 = Image.open(BytesIO(img_data1))
        img2 = Image.open(BytesIO(img_data2))

        # Create a new image with the target item size
        merged_img = Image.new('RGBA', (item_width, item_height))

        # Paste both images on the new canvas at their original sizes and positions
        # Adjust positioning logic if needed (assuming top-left corner for both images here)
        merged_img.paste(img1, (0, 0), img1)  # Use img1 as a mask to handle transparency
        merged_img.paste(img2, (0, 0), img2)  # Overlay img2 at the same position

        # Save the merged image to a buffer
        buffer = BytesIO()
        merged_img.save(buffer, format="PNG")
        merged_img_data = buffer.getvalue()

        # Encode the merged image back to a base64 string
        merged_data_url = "data:image/png;base64," + base64.b64encode(merged_img_data).decode()

        return merged_data_url

    def _merge_new_annotations(self, annotations_batch):
        """
        Merge the new binary annotations with the existing annotations
        :param annotations_batch: list of list of annotation. each batch with size self._upload_batch_size
        :return: merged_annotations_batch: list of list of annotation. each batch with size self._upload_batch_size
        """
        for annotations in annotations_batch:
            for annotation in annotations:
                if annotation['type'] == 'binary' and not annotation.get('clean', False):
                    to_merge = [a for a in annotations if
                                not a.get('clean', False) and a.get("metadata", {}).get('system', {}).get('objectId',
                                                                                                          None) ==
                                annotation.get("metadata", {}).get('system', {}).get('objectId', None) and a['label'] ==
                                annotation['label']]
                    if len(to_merge) == 0:
                        # no annotation to merge with
                        continue
                    for a in to_merge:
                        if a['coordinates'] == annotation['coordinates']:
                            continue
                        merged_data_url = self._merge_binary_annotations(a['coordinates'], annotation['coordinates'],
                                                                         self.item.width, self.item.height)
                        annotation['coordinates'] = merged_data_url
                        a['clean'] = True
        return [[a for a in annotations if not a.get('clean', False)] for annotations in annotations_batch]

    def _merge_to_exits_annotations(self, annotations_batch):
        filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION, field='type', values='binary')
        filters.add(field='itemId', values=self.item.id, method=entities.FiltersMethod.AND)
        exist_annotations = self.list(filters=filters).annotations or list()
        to_delete = list()
        for annotations in annotations_batch:
            for ann in annotations:
                if ann['type'] == 'binary':
                    to_merge = [a for a in exist_annotations if
                                a.object_id == ann.get("metadata", {}).get('system', {}).get('objectId',
                                                                                             None) and a.label == ann[
                                    'label']]
                    if len(to_merge) == 0:
                        # no annotation to merge with
                        continue
                    if to_merge[0].coordinates == ann['coordinates']:
                        # same annotation
                        continue
                    if len(to_merge) > 1:
                        raise exceptions.PlatformException('400', 'Multiple annotations with the same label')
                    # merge
                    exist_annotations.remove(to_merge[0])
                    merged_data_url = self._merge_binary_annotations(to_merge[0].coordinates, ann['coordinates'],
                                                                     self.item.width, self.item.height)
                    json_ann = to_merge[0].to_json()
                    json_ann['coordinates'] = merged_data_url
                    suc, response = self._update_annotation_req(annotation_json=json_ann,
                                                                system_metadata=True,
                                                                annotation_id=to_merge[0].id)
                    if not suc:
                        raise exceptions.PlatformException(response)
                    if suc:
                        result = entities.Annotation.from_json(_json=response.json(),
                                                               annotations=self,
                                                               dataset=self._dataset,
                                                               item=self._item)
                        exist_annotations.append(result)
                    to_delete.append(ann)
        if len(to_delete) > 0:
            annotations_batch = [[a for a in annotations if a not in to_delete] for annotations in annotations_batch]

        return annotations_batch

    def _upload_single_batch(self, annotation_batch):
        try:
            suc, response = self._client_api.gen_request(req_type='post',
                                                         path='/items/{}/annotations'.format(self.item.id),
                                                         json_req=annotation_batch)
            if suc:
                return_annotations = response.json()
                if not isinstance(return_annotations, list):
                    return_annotations = [return_annotations]
            else:
                raise exceptions.PlatformException(response)

            status = True
            result = return_annotations
        except Exception:
            status = False
            result = traceback.format_exc()

        return status, result

    def _upload_annotations_batches(self, annotation_batches):
        if len(annotation_batches) == 1:
            # no need for threads
            status, result = self._upload_single_batch(annotation_batch=annotation_batches[0])
            if status is False:
                logger.error(result)
                logger.error('Annotation/s uploaded with errors')
                # TODO need to raise errors?
            uploaded_annotations = result
        else:
            # threading
            pool = self._client_api.thread_pools(pool_name='annotation.upload')
            jobs = [None for _ in range(len(annotation_batches))]
            for i_ann, annotations_batch in enumerate(annotation_batches):
                jobs[i_ann] = pool.submit(self._upload_single_batch,
                                          annotation_batch=annotations_batch)
            # get all results
            results = [j.result() for j in jobs]
            uploaded_annotations = [ann for ann_list in results for ann in ann_list[1] if ann_list[0] is True]
            out_errors = [r[1] for r in results if r[0] is False]
            if len(out_errors) != 0:
                logger.error(out_errors)
                logger.error('Annotation/s uploaded with errors')
                # TODO need to raise errors?
        logger.info('Annotation/s uploaded successfully. num: {}'.format(len(uploaded_annotations)))
        return uploaded_annotations

    async def _async_upload_annotations(self, annotations, merge=False):
        """
        Async function to run from the uploader. will use asyncio to not break the async
        :param annotations: list of all annotations
        :param merge: bool - merge the new binary annotations with the existing annotations
        :return:
        """
        async with self._client_api.event_loop.semaphore('annotations.upload'):
            annotation_batch = self._create_batches_for_upload(annotations=annotations, merge=merge)
            output_annotations = list()
            for annotations_list in annotation_batch:
                success, response = await self._client_api.gen_async_request(req_type='post',
                                                                             path='/items/{}/annotations'
                                                                             .format(self.item.id),
                                                                             json_req=annotations_list)
                if success:
                    return_annotations = response.json()
                    if not isinstance(return_annotations, list):
                        return_annotations = [return_annotations]
                    output_annotations.extend(return_annotations)
                else:
                    if len(output_annotations) > 0:
                        logger.warning("Only {} annotations from {} annotations have been uploaded".
                                       format(len(output_annotations), len(annotations)))
                    raise exceptions.PlatformException(response)

            result = entities.AnnotationCollection.from_json(_json=output_annotations, item=self.item)
            return result

    @_api_reference.add(path='/items/{itemId}/annotations', method='post')
    def upload(self, annotations, merge=False) -> entities.AnnotationCollection:
        """
        Upload a new annotation/annotations. You must first create the annotation using the annotation *builder* method.

        **Prerequisites**: Any user can upload annotations.

        :param List[dtlpy.entities.annotation.Annotation] or dtlpy.entities.annotation.Annotation annotations: list or
        single annotation of type Annotation
        :param bool merge: optional - merge the new binary annotations with the existing annotations
        :return: list of annotation objects
        :rtype: entities.AnnotationCollection

        **Example**:

        .. code-block:: python

            annotations = item.annotations.upload(annotations='builder')
        """
        # make list if not list
        if isinstance(annotations, entities.AnnotationCollection):
            # get the annotation from a collection
            annotations = annotations.annotations
        elif isinstance(annotations, str) and os.path.isfile(annotations):
            # load annotation filepath and get list of annotations
            with open(annotations, 'r', encoding="utf8") as f:
                annotations = json.load(f)
            annotations = annotations.get('annotations', [])
            # annotations = entities.AnnotationCollection.from_json_file(filepath=annotations).annotations
        elif isinstance(annotations, entities.Annotation) or isinstance(annotations, dict):
            # convert the single Annotation to a list
            annotations = [annotations]
        elif isinstance(annotations, list):
            pass
        else:
            exceptions.PlatformException(error='400',
                                         message='Unknown annotation format. type: {}'.format(type(annotations)))
        if len(annotations) == 0:
            logger.warning('Annotation upload receives 0 annotations. Not doing anything')
            out_annotations = list()
        else:
            annotation_batches = self._create_batches_for_upload(annotations=annotations, merge=merge)
            out_annotations = self._upload_annotations_batches(annotation_batches=annotation_batches)
        out_annotations = entities.AnnotationCollection.from_json(_json=out_annotations,
                                                                  item=self.item)
        return out_annotations

    def update_status(self,
                      annotation: entities.Annotation = None,
                      annotation_id: str = None,
                      status: entities.AnnotationStatus = entities.AnnotationStatus.ISSUE
                      ) -> entities.Annotation:
        """
        Set status on annotation.

        **Prerequisites**: You must have an item that has been annotated. You must have the role of an *owner* or
        *developer* or be assigned a task that includes that item as an *annotation manager*.

        :param dtlpy.entities.annotation.Annotation annotation: Annotation object
        :param str annotation_id: optional - annotation id to set status
        :param str status: can be AnnotationStatus.ISSUE, APPROVED, REVIEW, CLEAR
        :return: Annotation object
        :rtype: dtlpy.entities.annotation.Annotation

        **Example**:

        .. code-block:: python

            annotation = item.annotations.update_status(annotation_id='annotation_id', status=dl.AnnotationStatus.ISSUE)
        """
        if annotation is None:
            if annotation_id is None:
                raise ValueError('must input on of "annotation" or "annotation_id"')
            annotation = self.get(annotation_id=annotation_id)
        if status not in list(entities.AnnotationStatus):
            raise ValueError('status must be on of: {}'.format(', '.join(list(entities.AnnotationStatus))))
        annotation.status = status
        return annotation.update(system_metadata=True)

    def builder(self):
        """
        Create Annotation collection.

        **Prerequisites**: You must have an item to be annotated. You must have the role of an *owner* or *developer*
         or be assigned a task that includes that item as an *annotation manager* or *annotator*.

        :return: Annotation collection object
        :rtype: dtlpy.entities.annotation_collection.AnnotationCollection

        **Example**:

        .. code-block:: python

            annotation_collection = item.annotations.builder()
        """
        return entities.AnnotationCollection(item=self.item)

    def task_scores(self, annotation_id: str, task_id: str, page_offset: int = 0, page_size: int = 100):
        """
        Get annotation scores in a task

        **Prerequisites**: You must be able to read the task

        :param str annotation_id: The id of the annotation
        :param str task_id: The id of the task
        :param int page_offset: starting page
        :param int page_size: size of page
        :return: json response
        :rtype: dict
        """
        if annotation_id is None:
            raise exceptions.PlatformException('400', 'annotation_id must be provided')
        if task_id is None:
            raise exceptions.PlatformException('400', 'task_id must be provided')

        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/scores/tasks/{}/annotations/{}?page={}&pageSize={}'
                                                         .format(task_id, annotation_id, page_offset, page_size))
        if success:
            return response.json()
        else:
            raise exceptions.PlatformException(response)

    ##################
    # async function #
    ##################



import sys

from jinja2 import Environment, PackageLoader
from multiprocessing.pool import ThreadPool
from multiprocessing import Lock
from .base_package_runner import Progress
from .. import exceptions, entities
import xml.etree.ElementTree as Et
from ..services import Reporter
from itertools import groupby
from PIL import Image
import numpy as np
import mimetypes
import traceback
import logging
import json
import copy
import tqdm
import os

logger = logging.getLogger(name='dtlpy')


class AnnotationFormat:
    YOLO = 'yolo'
    COCO = 'coco'
    VOC = 'voc'
    DATALOOP = 'dataloop'


class COCOUtils:

    @staticmethod
    def binary_mask_to_rle_encode(binary_mask):
        try:
            import pycocotools.mask as coco_utils_mask
        except ModuleNotFoundError:
            raise Exception('To use this functionality please install pycocotools:  "pip install pycocotools"')
        fortran_ground_truth_binary_mask = np.asfortranarray(binary_mask.astype(np.uint8))
        encoded_ground_truth = coco_utils_mask.encode(fortran_ground_truth_binary_mask)
        encoded_ground_truth['counts'] = encoded_ground_truth['counts'].decode()
        return encoded_ground_truth

    @staticmethod
    def binary_mask_to_rle(binary_mask, height, width):
        rle = {'counts': [], 'size': [height, width]}
        counts = rle.get('counts')
        for i, (value, elements) in enumerate(groupby(binary_mask.ravel(order='F'))):
            if i == 0 and value == 1:
                counts.append(0)
            counts.append(len(list(elements)))
        return rle

    @staticmethod
    def polygon_to_rle(geo, height, width):
        segmentation = [float(n) for n in geo.flatten()]
        area = np.sum(entities.Segmentation.from_polygon(geo=geo, label=None, shape=(height, width)).geo > 0)
        return [segmentation], int(area)

    @staticmethod
    def rle_to_binary_mask(rle):
        rows, cols = rle['size']
        rle_numbers = rle['counts']
        if isinstance(rle_numbers, list):
            if len(rle_numbers) % 2 != 0:
                rle_numbers.append(0)

            rle_pairs = np.array(rle_numbers).reshape(-1, 2)
            img = np.zeros(rows * cols, dtype=np.uint8)
            index = 0
            for i, length in rle_pairs:
                index += i
                img[index:index + length] = 1
                index += length
            img = img.reshape(cols, rows)
            return img.T
        else:
            try:
                import pycocotools.mask as coco_utils_mask
            except ModuleNotFoundError:
                raise Exception('To use this functionality please install pycocotools:  "pip install pycocotools"')
            img = coco_utils_mask.decode(rle)
            return img

    @staticmethod
    def rle_to_binary_polygon(segmentation):
        return [segmentation[x:x + 2] for x in range(0, len(segmentation), 2)]


class Converter:
    """
    Annotation Converter
    """

    def __init__(self, concurrency=6, return_error_filepath=False):
        self.known_formats = [AnnotationFormat.YOLO,
                              AnnotationFormat.COCO,
                              AnnotationFormat.VOC,
                              AnnotationFormat.DATALOOP]
        self.converter_dict = {
            AnnotationFormat.YOLO: {"from": self.from_yolo, "to": self.to_yolo},
            AnnotationFormat.COCO: {"from": self.from_coco, "to": self.to_coco},
            AnnotationFormat.VOC: {"from": self.from_voc, "to": self.to_voc},
        }
        self.dataset = None
        self.save_to_format = None
        self.xml_template_path = 'voc_annotation_template.xml'
        self.labels = dict()
        self._only_bbox = False
        self._progress = None
        self._progress_update_frequency = 5
        self._update_agent_progress = False
        self._progress_checkpoint = 0
        self._checkpoint_lock = Lock()
        self.remote_items = None
        self.concurrency = concurrency
        self.return_error_filepath = return_error_filepath

    def attach_agent_progress(self, progress: Progress, progress_update_frequency: int = None):
        """
        Attach agent progress.

        :param Progress progress: the progress object that follows the work
        :param int progress_update_frequency: progress update frequency in percentages
        """
        self._progress = progress
        self._progress_update_frequency = progress_update_frequency if progress_update_frequency is not None \
            else self._progress_update_frequency
        self._update_agent_progress = True

    @property
    def _update_progress_active(self):
        return self._progress is not None and self._update_agent_progress and isinstance(
            self._progress_update_frequency, int)

    def __update_progress(self, total, of_total):
        if self._update_progress_active:
            try:
                progress = int((of_total / total) * 100)
                if progress > self._progress_checkpoint and (progress % self._progress_update_frequency == 0):
                    self._progress.update(progress=progress)
                    with self._checkpoint_lock:
                        self._progress_checkpoint = progress
            except Exception:
                logger.warning('[Converter] Failed to update agent progress.')

    def _get_labels(self):
        self.labels = dict()
        if self.dataset:
            labels = list(self.dataset.instance_map.keys())
            labels.sort()
            self.labels = {label: i for i, label in enumerate(labels)}

    def convert_dataset(
            self,
            dataset,
            to_format: str,
            local_path: str,
            conversion_func=None,
            filters=None,
            annotation_filter=None
    ):
        """
        Convert entire dataset.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.dataet.Dataset dataset: dataset entity
        :param str to_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param str local_path: path to save the result to
        :param Callable conversion_func: Custom conversion service
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filter parameters
        :param dtlpy.entities.filters.Filters annotation_filter: Filter entity
        :return: the error log file path if there are errors and the coco json if the format is coco
        """
        if to_format.lower() == AnnotationFormat.COCO:
            coco_json, has_errors, log_filepath = self.__convert_dataset_to_coco(
                dataset=dataset,
                local_path=local_path,
                filters=filters,
                annotation_filter=annotation_filter
            )

            if self.return_error_filepath:
                return coco_json, has_errors, log_filepath
            else:
                return coco_json

        assert isinstance(dataset, entities.Dataset)
        self.dataset = dataset

        # download annotations
        if annotation_filter is None:
            logger.info('Downloading annotations...')
            dataset.download_annotations(local_path=local_path, overwrite=True, filters=filters)
            logger.info('Annotations downloaded')
        local_annotations_path = os.path.join(local_path, "json")
        output_annotations_path = os.path.join(local_path, to_format)
        pool = ThreadPool(processes=self.concurrency)
        i_item = 0
        pages = dataset.items.list(filters=filters)

        # if yolo - create labels file
        if to_format == AnnotationFormat.YOLO:
            self._get_labels()
            with open('{}/{}.names'.format(local_path, dataset.name), 'w') as fp:
                labels = list(self.labels.keys())
                labels.sort()
                for label in labels:
                    fp.write("{}\n".format(label))

        pbar = tqdm.tqdm(total=pages.items_count, disable=dataset._client_api.verbose.disable_progress_bar_convert_annotations,
                         file=sys.stdout, desc='Convert Annotations')
        reporter = Reporter(
            num_workers=pages.items_count,
            resource=Reporter.CONVERTER,
            print_error_logs=self.dataset._client_api.verbose.print_error_logs,
            client_api=self.dataset._client_api
        )
        for page in pages:
            for item in page:
                # create input annotations json
                in_filepath = os.path.join(local_annotations_path, item.filename[1:])
                name, ext = os.path.splitext(in_filepath)
                in_filepath = name + '.json'

                save_to = os.path.dirname(in_filepath.replace(local_annotations_path, output_annotations_path))

                if not os.path.isdir(save_to):
                    os.makedirs(save_to, exist_ok=True)

                pool.apply_async(
                    func=self.__save_filtered_annotations_and_convert,
                    kwds={
                        "to_format": to_format,
                        "from_format": AnnotationFormat.DATALOOP,
                        "file_path": in_filepath,
                        "save_locally": True,
                        "save_to": save_to,
                        'conversion_func': conversion_func,
                        'item': item,
                        'pbar': pbar,
                        'filters': annotation_filter,
                        'reporter': reporter,
                        'i_item': i_item
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()

        log_filepath = None

        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            if log_filepath is not None:
                logger.warning(
                    'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted: {}'.format(reporter.status_count('success')))
        logger.info('Total skipped: {}'.format(reporter.status_count('skip')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

        if self.return_error_filepath:
            return reporter.has_errors, log_filepath

    def __save_filtered_annotations_and_convert(self, item: entities.Item, filters, to_format, from_format, file_path,
                                                save_locally=False,
                                                save_to=None, conversion_func=None,
                                                pbar=None, **kwargs):
        reporter = kwargs.get('reporter', None)
        i_item = kwargs.get('i_item', None)

        try:
            if item.annotated and item.type != 'dir':
                if filters is not None:
                    assert filters.resource == entities.FiltersResource.ANNOTATION
                    copy_filters = copy.deepcopy(filters)
                    copy_filters.add(field='itemId', values=item.id, method='and')
                    annotations_page = item.dataset.items.list(filters=copy_filters)
                    annotations = item.annotations.builder()
                    for page in annotations_page:
                        for annotation in page:
                            annotations.annotations.append(annotation)

                    if not os.path.isdir(os.path.dirname(file_path)):
                        os.makedirs(os.path.dirname(file_path))
                    with open(file_path, 'w') as f:
                        json.dump(annotations.to_json(), f, indent=2)

                annotations_list, errors = self.convert_file(
                    item=item,
                    to_format=to_format,
                    from_format=from_format,
                    file_path=file_path,
                    save_locally=save_locally,
                    save_to=save_to,
                    conversion_func=conversion_func,
                    pbar=pbar
                )

                if errors:
                    raise Exception('Partial conversion: \n{}'.format(errors))

                if reporter is not None and i_item is not None:
                    reporter.set_index(status='success', success=True, ref=item.id)
            else:
                if reporter is not None and i_item is not None:
                    reporter.set_index(ref=item.id, status='skip', success=True)
            if reporter is not None:
                self.__update_progress(total=reporter.num_workers, of_total=i_item)
        except Exception:
            if reporter is not None and i_item is not None:
                reporter.set_index(status='failed', success=False, error=traceback.format_exc(),
                                   ref=item.id)

    @staticmethod
    def __gen_coco_categories(instance_map, recipe):
        categories = list()
        last_id = 0
        for label, label_id in instance_map.items():
            label_name, sup = label.split('.')[-1], '.'.join(label.split('.')[0:-1])
            category = {'id': label_id, 'name': label_name}
            last_id = max(last_id, label_id)
            if sup:
                category['supercategory'] = sup
            categories.append(category)

        # add keypoint category
        collection_templates = list()
        if 'system' in recipe.metadata and 'collectionTemplates' in recipe.metadata['system']:
            collection_templates = recipe.metadata['system']['collectionTemplates']

        for template in collection_templates:
            last_id += 1
            order_dict = {key: i for i, key in enumerate(template['order'])}
            skeleton = list()
            for pair in template['arcs']:
                skeleton.append([order_dict[pair[0]], order_dict[pair[1]]])
            category = {'id': last_id,
                        'name': template['name'],
                        'templateId': template['id'],
                        'keypoints': template['order'],
                        'skeleton': skeleton}
            instance_map[template['name']] = last_id
            categories.append(category)
        return categories

    def __convert_dataset_to_coco(self, dataset: entities.Dataset, local_path, filters=None, annotation_filter=None):
        pages = dataset.items.list(filters=filters)
        logger.info('items count: {}'.format(pages.items_count))
        dataset.download_annotations(local_path=local_path, filters=filters, annotation_filters=annotation_filter)
        path_to_dataloop_annotations_dir = os.path.join(local_path, 'json')
        label_to_id = dataset.instance_map
        recipe = dataset._get_recipe()
        categories = self.__gen_coco_categories(instance_map=label_to_id, recipe=recipe)
        images = [None for _ in range(pages.items_count)]
        converted_annotations = [None for _ in range(pages.items_count)]
        item_id_counter = 0
        pool = ThreadPool(processes=self.concurrency)
        pbar = tqdm.tqdm(total=pages.items_count, disable=dataset._client_api.verbose.disable_progress_bar_convert_annotations,
                         file=sys.stdout, desc='Convert Annotations')
        reporter = Reporter(
            num_workers=pages.items_count,
            resource=Reporter.CONVERTER,
            print_error_logs=dataset._client_api.verbose.print_error_logs,
            client_api=dataset._client_api
        )
        for page in pages:
            for item in page:
                try:
                    pool.apply_async(func=self.__single_item_to_coco,
                                     kwds={
                                         'item': item,
                                         'images': images,
                                         'path_to_dataloop_annotations_dir': path_to_dataloop_annotations_dir,
                                         'item_id': item_id_counter,
                                         'reporter': reporter,
                                         'converted_annotations': converted_annotations,
                                         'annotation_filter': annotation_filter,
                                         'label_to_id': label_to_id,
                                         'categories': categories,
                                         'pbar': pbar
                                     })
                    item_id_counter += 1
                except Exception as e:
                    logger.error('Failed to convert item: {}'.format(item.id))

        pool.close()
        pool.join()
        pool.terminate()

        total_converted_annotations = list()
        for ls in converted_annotations:
            if ls is not None:
                total_converted_annotations += ls

        for i_ann, ann in enumerate(total_converted_annotations):
            ann['id'] = i_ann

        info = {
            'description': dataset.name
        }

        coco_json = {'images': [image for image in images if image is not None],
                     'info': info,
                     'annotations': total_converted_annotations,
                     'categories': categories}

        with open(os.path.join(local_path, 'coco.json'), 'w+', encoding='utf-8') as f:
            json.dump(coco_json, f, ensure_ascii=False)

        log_filepath = None
        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            if log_filepath is not None:
                logger.warning(
                    'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted: {}'.format(reporter.success_count))
        logger.info('Total failed: {}'.format(reporter.failure_count))
        logger.info('Total skipped: {}'.format(reporter.status_count('skip')))
        if reporter.success_count + reporter.status_count('skip') + reporter.failure_count != pages.items_count:
            raise ValueError('Not all items were processed')

        return coco_json, reporter.has_errors, log_filepath

    @staticmethod
    def __get_item_shape(item: entities.Item = None, local_path: str = None):
        if isinstance(item, entities.Item) and (item.width is None or item.height is None):
            try:
                img = Image.open(item.download(save_locally=False)) if local_path is None else Image.open(local_path)
                item.height = img.height
                item.width = img.width
            except Exception:
                pass
        return item

    def __add_item_converted_annotation(self, item, annotation, label_to_id, item_id,
                                        i_annotation, item_converted_annotations):
        try:
            ann = self.to_coco(annotation=annotation, item=item)
            ann['category_id'] = label_to_id[annotation.label]
            ann['image_id'] = item_id
            ann['id'] = int('{}{}'.format(item_id, i_annotation))

            item_converted_annotations.append(ann)
        except Exception:
            err = 'Error converting annotation: \n' \
                  'Item: {}, annotation: {} - ' \
                  'fail to convert some of the annotation\n{}'.format(item_id,
                                                                      annotation.id,
                                                                      traceback.format_exc())
            item_converted_annotations.append(err)

    def __coco_handle_pose_annotations(self, item, item_id, pose_annotations, point_annotations,
                                       categories, label_to_id, item_converted_annotations):
        # link points to pose and convert it
        for pose in pose_annotations:
            pose_idx = pose[1]
            pose = pose[0]
            pose_category = None
            for category in categories:
                if pose.coordinates.get('templateId', "") == category.get('templateId', None):
                    pose_category = category
                    continue
            if pose_category is None:
                err = 'Error converting annotation: \n' \
                      'Item: {}, annotation: {} - ' \
                      'Pose annotation without known template\n{}'.format(item_id,
                                                                          pose.id,
                                                                          traceback.format_exc())
                item_converted_annotations.append(err)
                continue
            if pose.id not in point_annotations or (pose.id in point_annotations and
                                                    len(point_annotations[pose.id]) != len(pose_category['keypoints'])):
                err = 'Error converting annotation: \n' \
                      'Item: {}, annotation: {} - ' \
                      'Pose annotation has {} children ' \
                      'while it template has {} points\n{}'.format(item_id,
                                                                   pose.id,
                                                                   len(point_annotations[pose.id]),
                                                                   len(pose_category['keypoints']),
                                                                   traceback.format_exc())
                item_converted_annotations.append(err)
                continue
            # verify points labels are unique
            if len(point_annotations[pose.id]) != len(set([ann.label for ann in point_annotations[pose.id]])):
                err = 'Error converting annotation: \n' \
                      'Item: {}, annotation: {} - Pose annotation ' \
                      'does not have unique children points\n{}'.format(item_id,
                                                                        pose.id,
                                                                        traceback.format_exc())
                item_converted_annotations.append(err)
                continue

            ordered_points = list()
            for pose_point in pose_category['keypoints']:
                for point_annotation in point_annotations[pose.id]:
                    if point_annotation.label == pose_point:
                        ordered_points.append(point_annotation)
                        break
            pose.annotation_definition.points = ordered_points

            self.__add_item_converted_annotation(item=item, annotation=pose,
                                                 label_to_id=label_to_id,
                                                 item_id=item_id, i_annotation=pose_idx,
                                                 item_converted_annotations=item_converted_annotations)

    def __single_item_to_coco(self, item: entities.Item,
                              images,
                              path_to_dataloop_annotations_dir,
                              item_id,
                              converted_annotations,
                              annotation_filter,
                              label_to_id,
                              reporter,
                              categories,
                              pbar=None):
        try:
            if item.type != 'dir':
                item = Converter.__get_item_shape(item=item)
                filepath = item.filename[1:] if item.filename.startswith('/') else item.filename
                images[item_id] = {'file_name': os.path.normpath(filepath),
                                   'id': item_id,
                                   'width': item.width,
                                   'height': item.height
                                   }
                if annotation_filter is None:
                    try:
                        filename, ext = os.path.splitext(item.filename)
                        filename = '{}.json'.format(filename[1:])
                        with open(os.path.join(path_to_dataloop_annotations_dir, filename), 'r', encoding="utf8") as f:
                            annotations = json.load(f)['annotations']
                        annotations = entities.AnnotationCollection.from_json(annotations)
                    except Exception:
                        annotations = item.annotations.list()
                else:
                    copy_filters = copy.deepcopy(annotation_filter)
                    copy_filters.add(field='itemId', values=item.id, method='and')
                    annotations_page = item.dataset.items.list(filters=copy_filters)
                    annotations = item.annotations.builder()
                    for page in annotations_page:
                        for annotation in page:
                            annotations.annotations.append(annotation)

                item_converted_annotations = list()
                point_annotations = dict()
                pose_annotations = list()

                for i_annotation, annotation in enumerate(annotations.annotations):
                    if annotation.type == 'point' and annotation.parent_id is not None:
                        if annotation.parent_id not in point_annotations:
                            point_annotations[annotation.parent_id] = list()
                        point_annotations[annotation.parent_id].append(annotation)
                        continue
                    if annotation.type == 'pose':
                        pose_annotations.append([annotation, i_annotation])
                        continue
                    self.__add_item_converted_annotation(item=item, annotation=annotation,
                                                         label_to_id=label_to_id,
                                                         item_id=item_id, i_annotation=i_annotation,
                                                         item_converted_annotations=item_converted_annotations)

                self.__coco_handle_pose_annotations(item=item, item_id=item_id,
                                                    pose_annotations=pose_annotations,
                                                    point_annotations=point_annotations,
                                                    categories=categories,
                                                    label_to_id=label_to_id,
                                                    item_converted_annotations=item_converted_annotations)

                success, errors = self._sort_annotations(annotations=item_converted_annotations)
                converted_annotations[item_id] = success
                if errors:
                    reporter.set_index(ref=item.id, status='failed', success=False,
                                       error=errors)
                else:
                    reporter.set_index(ref=item.id, status='success', success=True)
            else:
                reporter.set_index(ref=item.id, status='skipped', success=True)
        except Exception:
            reporter.set_index(ref=item.id, status='failed', success=False,
                               error=traceback.format_exc())
            logger.debug('Error converting item: {}\n{}'.format(item.id, traceback.format_exc()))
            raise Exception('Error converting item: {}\n{}'.format(item.id, traceback.format_exc()))
        if pbar is not None:
            pbar.update()

        if reporter is not None:
            self.__update_progress(total=reporter.num_workers, of_total=item_id)

    def _upload_coco_labels(self, coco_json):
        labels = coco_json.get('categories', None)
        upload_labels = dict()
        for label in labels:
            if 'supercategory' in label and label['supercategory'] is not None:
                if label['supercategory'] not in upload_labels:
                    upload_labels[label['supercategory']] = entities.Label(tag=label['supercategory'])
                upload_labels[label['supercategory']].children.append(entities.Label(tag=label['name']))
                tag = '{}.{}'.format(label['supercategory'], label['name'])
            else:
                tag = label['name']
                upload_labels[label['name']] = entities.Label(tag=tag)
            self.labels[tag] = label['id']

        return upload_labels

    def _upload_coco_dataset(self, local_items_path, local_annotations_path, only_bbox=False, remote_items=False):
        logger.info('loading annotations json...')
        with open(local_annotations_path, 'r', encoding="utf8") as f:
            coco_json = json.load(f)

        labels_tags_tree = self._upload_coco_labels(coco_json=coco_json)
        try:
            logger.info('Uploading labels to dataset')
            self.dataset.add_labels(list(labels_tags_tree.values()))
        except Exception:
            logger.warning('Failed to upload labels to dataset, please add manually')

        image_annotations = dict()
        image_name_id = dict()
        for image in coco_json['images']:
            image_metadata = image
            image_annotations[image['file_name']] = {
                'id': image['id'],
                'metadata': image_metadata,
                'annotations': list()
            }
            image_name_id[image['id']] = image['file_name']
        for ann in coco_json['annotations']:
            image_annotations[image_name_id[ann['image_id']]]['annotations'].append(ann)

        if remote_items:
            return self._upload_annotations(local_annotations_path=image_annotations,
                                            from_format=AnnotationFormat.COCO,
                                            only_bbox=only_bbox)
        else:
            return self._upload_directory(local_items_path=local_items_path,
                                          local_annotations_path=image_annotations,
                                          from_format=AnnotationFormat.COCO,
                                          only_bbox=only_bbox)

    def _read_labels(self, labels_file_path):
        if labels_file_path:
            with open(labels_file_path, 'r') as fp:
                labels = [line.strip() for line in fp.readlines()]
                self.dataset.add_labels(label_list=labels)
                self.labels = {label: i_label for i_label, label in enumerate(labels)}
        else:
            logger.warning('No labels file path provided (.names), skipping labels upload')
            self._get_labels()

    def _upload_yolo_dataset(self, local_items_path, local_annotations_path, labels_file_path, remote_items=False):
        self._read_labels(labels_file_path=labels_file_path)
        if remote_items:
            return self._upload_annotations(local_annotations_path=local_annotations_path,
                                            from_format=AnnotationFormat.YOLO)
        else:
            return self._upload_directory(local_items_path=local_items_path,
                                          local_annotations_path=local_annotations_path,
                                          from_format=AnnotationFormat.YOLO)

    def _upload_voc_dataset(self, local_items_path, local_annotations_path, remote_items=False, **_):
        # TODO - implement VOC annotations upload
        logger.warning('labels upload from VOC dataset is not implemented, please upload labels manually')

        if remote_items:
            return self._upload_annotations(local_annotations_path=local_annotations_path,
                                            from_format=AnnotationFormat.VOC)
        else:
            return self._upload_directory(local_items_path=local_items_path,
                                          local_annotations_path=local_annotations_path,
                                          from_format=AnnotationFormat.VOC)

    @staticmethod
    def _find_yolo_voc_item_annotations(local_annotations_path: str, item: entities.Item, from_format: str):
        found = False
        metadata = None

        extension = '.txt' if from_format == AnnotationFormat.YOLO else '.xml'
        filename, _ = os.path.splitext(item.filename)
        annotations_filepath = os.path.join(local_annotations_path, filename[1:] + extension)
        if os.path.isfile(annotations_filepath):
            found = True

        return found, annotations_filepath, metadata

    @staticmethod
    def _find_coco_item_annotations(local_annotations_path: dict, item: entities.Item):
        found = False
        ann_dict = None
        filename = item.filename[1:] if item.filename.startswith('/') else item.filename
        filename = os.path.normpath(filename)
        if filename in local_annotations_path:
            ann_dict = local_annotations_path[filename]
            found = True
        elif item.name in local_annotations_path:
            ann_dict = local_annotations_path[item.name]
            found = True
        metadata = ann_dict.get('metadata', None) if found else None
        return found, ann_dict, metadata

    def _upload_annotations(self, local_annotations_path, from_format, **kwargs):
        self._only_bbox = kwargs.get('only_bbox', False)
        file_count = self.remote_items.items_count
        reporter = Reporter(
            num_workers=file_count,
            resource=Reporter.CONVERTER,
            print_error_logs=self.dataset._client_api.verbose.print_error_logs,
            client_api=self.dataset._client_api
        )

        pbar = tqdm.tqdm(total=file_count, desc='Upload Annotations')
        pool = ThreadPool(processes=self.concurrency)
        i_item = 0

        for page in self.remote_items:
            for item in page:
                if from_format == AnnotationFormat.COCO:
                    found, ann_filepath, metadata = self._find_coco_item_annotations(
                        local_annotations_path=local_annotations_path,
                        item=item
                    )
                elif from_format in [AnnotationFormat.YOLO, AnnotationFormat.VOC]:
                    found, ann_filepath, metadata = self._find_yolo_voc_item_annotations(
                        local_annotations_path=local_annotations_path,
                        item=item,
                        from_format=from_format
                    )
                else:
                    raise exceptions.PlatformException('400', 'Unknown annotation format: {}'.format(from_format))
                if not found:
                    pbar.update()
                    reporter.set_index(ref=item.filename, status='skip', success=False,
                                       error='Cannot find annotations for item')
                    i_item += 1
                    continue
                pool.apply_async(
                    func=self._upload_item_and_convert,
                    kwds={
                        "from_format": from_format,
                        "item_path": item,
                        "ann_path": ann_filepath,
                        'conversion_func': None,
                        "reporter": reporter,
                        'i_item': i_item,
                        'pbar': pbar,
                        'metadata': metadata
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()

        log_filepath = None
        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            if log_filepath is not None:
                logger.warning(
                    'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted and uploaded: {}'.format(reporter.status_count('success')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

        if self.return_error_filepath:
            return reporter.has_errors, log_filepath

    def _upload_directory(self, local_items_path, local_annotations_path, from_format, conversion_func=None, **kwargs):
        """
        Convert annotation files in entire directory.

        :param local_items_path:
        :param local_annotations_path:
        :param from_format:
        :param conversion_func:
        :return:
        """
        self._only_bbox = kwargs.get('only_bbox', False)
        file_count = sum(len([file for file in files if not file.endswith('.xml')]) for _, _, files in
                         os.walk(local_items_path))

        reporter = Reporter(
            num_workers=file_count,
            resource=Reporter.CONVERTER,
            print_error_logs=self.dataset._client_api.verbose.print_error_logs,
            client_api=self.dataset._client_api
        )

        pbar = tqdm.tqdm(total=file_count, desc='Upload Annotations')

        pool = ThreadPool(processes=self.concurrency)
        i_item = 0
        metadata = None
        for path, subdirs, files in os.walk(local_items_path):
            for name in files:
                ann_filepath = None
                if not os.path.isfile(os.path.join(path, name)):
                    continue
                item_filepath = os.path.join(path, name)
                prefix = os.path.relpath(path, local_items_path)
                coco_name = name
                if prefix != '.':
                    coco_name = os.path.join(prefix, name)
                else:
                    prefix = None
                if from_format == AnnotationFormat.COCO:
                    ann_filepath = local_annotations_path[coco_name]
                    metadata = {'user': local_annotations_path[coco_name]['metadata']}
                elif from_format == AnnotationFormat.VOC:
                    if name.endswith('.xml'):
                        continue
                    else:
                        ext = os.path.splitext(name)[-1]
                        try:
                            m = mimetypes.types_map[ext.lower()]
                        except Exception:
                            m = ''

                        if ext == '' or ext is None or 'image' not in m:
                            continue

                    ann_filepath = os.path.join(path, '.'.join(name.split('.')[0:-1] + ['xml'])).replace(
                        local_items_path, local_annotations_path)
                elif from_format == AnnotationFormat.YOLO:
                    ann_filepath = os.path.join(path, os.path.splitext(name)[0]) + '.txt'
                    ann_filepath = ann_filepath.replace(local_items_path, local_annotations_path)
                pool.apply_async(
                    func=self._upload_item_and_convert,
                    kwds={
                        "from_format": from_format,
                        "item_path": item_filepath,
                        "ann_path": ann_filepath,
                        'conversion_func': conversion_func,
                        "reporter": reporter,
                        'i_item': i_item,
                        'pbar': pbar,
                        'metadata': metadata,
                        'remote_path': prefix
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()

        log_filepath = None
        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            if log_filepath is not None:
                logger.warning(
                    'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted and uploaded: {}'.format(reporter.status_count('success')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

        if self.return_error_filepath:
            return reporter.has_errors, log_filepath

    def _upload_item_and_convert(self, item_path, ann_path, from_format, conversion_func=None, **kwargs):
        reporter = kwargs.get('reporter', None)
        i_item = kwargs.get('i_item', None)
        pbar = kwargs.get('pbar', None)
        metadata = kwargs.get('metadata', None)
        remote_path = kwargs.get('remote_path', None)
        report_ref = item_path
        try:
            if isinstance(item_path, entities.Item):
                item = item_path
                report_ref = item.filename
            else:
                item = self.dataset.items.upload(local_path=item_path, item_metadata=metadata, remote_path=remote_path)
                report_ref = item.filename
            if from_format == AnnotationFormat.YOLO:
                item = Converter.__get_item_shape(item=item, local_path=item_path)
            annotations_list, errors = self.convert_file(to_format=AnnotationFormat.DATALOOP,
                                                         from_format=from_format,
                                                         item=item,
                                                         file_path=ann_path,
                                                         save_locally=False,
                                                         conversion_func=conversion_func,
                                                         upload=True,
                                                         pbar=pbar)
            if errors:
                if reporter is not None and i_item is not None:
                    reporter.set_index(ref=report_ref,
                                       status='warning',
                                       success=False,
                                       error='partial annotations upload: \n{}'.format(errors))
            else:
                if reporter is not None and i_item is not None:
                    reporter.set_index(status='success',
                                       success=True,
                                       ref=report_ref)
            if reporter is not None:
                self.__update_progress(total=reporter.num_workers, of_total=i_item)
        except Exception:
            if reporter is not None and i_item is not None:
                reporter.set_index(status='failed',
                                   success=False,
                                   error=traceback.format_exc(),
                                   ref=report_ref)

    def upload_local_dataset(self,
                             from_format: AnnotationFormat,
                             dataset,
                             local_items_path: str = None,
                             local_labels_path: str = None,
                             local_annotations_path: str = None,
                             only_bbox: bool = False,
                             filters=None,
                             remote_items=None
                             ):
        """
        Convert and upload local dataset to dataloop platform.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str from_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param dtlpy.entities.dataset.Dataset dataset: dataset entity
        :param str local_items_path: path to items to upload
        :param str local_annotations_path: path to annotations to upload
        :param str local_labels_path: path to labels to upload
        :param bool only_bbox: only for coco datasets, if True upload only bbox
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filter parameters
        :param list remote_items: list of the items to upload
        :return: the error log file path if there are errors
        """

        if remote_items is None:
            remote_items = local_items_path is None

        if remote_items:
            logger.info('Getting remote items...')
            self.remote_items = dataset.items.list(filters=filters)

        self.dataset = dataset
        if from_format.lower() == AnnotationFormat.COCO:
            return self._upload_coco_dataset(
                local_items_path=local_items_path,
                local_annotations_path=local_annotations_path,
                only_bbox=only_bbox,
                remote_items=remote_items
            )
        if from_format.lower() == AnnotationFormat.YOLO:
            return self._upload_yolo_dataset(
                local_items_path=local_items_path,
                local_annotations_path=local_annotations_path,
                labels_file_path=local_labels_path,
                remote_items=remote_items
            )
        if from_format.lower() == AnnotationFormat.VOC:
            return self._upload_voc_dataset(
                local_items_path=local_items_path,
                local_annotations_path=local_annotations_path,
                labels_file_path=local_labels_path,
                remote_items=remote_items
            )

    def convert_directory(self,
                          local_path: str,
                          to_format: AnnotationFormat,
                          from_format: AnnotationFormat,
                          dataset,
                          conversion_func=None):
        """
        Convert annotation files in entire directory.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str local_path: path to the directory
        :param str to_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param str from_format: AnnotationFormat to convert from – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param dtlpy.entities.dataset.Dataset dataset: dataset entity
        :param Callable conversion_func: Custom conversion service
        :return: the error log file path if there are errors
        """
        file_count = sum(len(files) for _, _, files in os.walk(local_path))
        self.dataset = dataset
        reporter = Reporter(
            num_workers=file_count,
            resource=Reporter.CONVERTER,
            print_error_logs=self.dataset._client_api.verbose.print_error_logs,
            client_api=self.dataset._client_api
        )

        pool = ThreadPool(processes=self.concurrency)
        i_item = 0
        for path, subdirs, files in os.walk(local_path):
            for name in files:
                save_to = os.path.join(os.path.split(local_path)[0], to_format)
                save_to = path.replace(local_path, save_to)
                os.makedirs(save_to, exist_ok=True)
                file_path = os.path.join(path, name)
                pool.apply_async(
                    func=self._convert_and_report,
                    kwds={
                        "to_format": to_format,
                        "from_format": from_format,
                        "file_path": file_path,
                        "save_locally": True,
                        "save_to": save_to,
                        'conversion_func': conversion_func,
                        'reporter': reporter,
                        'i_item': i_item
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()

        log_filepath = None
        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            if log_filepath is not None:
                logger.warning(
                    'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted and uploaded: {}'.format(reporter.status_count('success')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

        if self.return_error_filepath:
            return reporter.has_errors, log_filepath

    def _convert_and_report(self, to_format, from_format, file_path, save_locally, save_to, conversion_func, reporter,
                            i_item):
        try:
            annotations_list, errors = self.convert_file(
                to_format=to_format,
                from_format=from_format,
                file_path=file_path,
                save_locally=save_locally,
                save_to=save_to,
                conversion_func=conversion_func
            )
            if errors:
                reporter.set_index(ref=file_path, status='warning', success=False,
                                   error='partial annotations upload')
            else:
                reporter.set_index(ref=file_path, status='success', success=True)
        except Exception:
            reporter.set_index(ref=file_path, status='failed', success=False,
                               error=traceback.format_exc())
            raise

    @staticmethod
    def _extract_annotations_from_file(file_path, from_format, item):
        item_id = None
        annotations = None
        if isinstance(file_path, dict):
            annotations = file_path['annotations']
        elif isinstance(file_path, str) and os.path.isfile(file_path):
            with open(file_path, "r") as f:
                if file_path.endswith(".json"):
                    annotations = json.load(f)
                    if from_format.lower() == AnnotationFormat.DATALOOP:
                        if item is None:
                            item_id = annotations.get('_id', annotations.get('id', None))
                        annotations = annotations["annotations"]
                elif from_format.lower() == AnnotationFormat.YOLO:
                    annotations = [[float(param.replace('\n', ''))
                                    for param in line.strip().split(' ')]
                                   for line in f.readlines()]
                elif file_path.endswith(".xml"):
                    annotations = Et.parse(f)
                    annotations = [e for e in annotations.iter('object')]
                else:
                    raise NotImplementedError('Unknown file format: {}'.format(file_path))
        else:
            raise NotImplementedError('Unknown file_path: {}'.format(file_path))

        return item_id, annotations

    def convert_file(self,
                     to_format: str,
                     from_format: str,
                     file_path: str,
                     save_locally: bool = False,
                     save_to: str = None,
                     conversion_func=None,
                     item=None,
                     pbar=None,
                     upload: bool = False,
                     **_):
        """
        Convert file containing annotations.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str to_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param str from_format: AnnotationFormat to convert from – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param str file_path: path of the file to convert
        :param tqdm pbar: tqdm object that follows the work (progress bar)
        :param bool upload: if True upload
        :param bool save_locally: If True, save locally
        :param str save_to: path to save the result to
        :param Callable conversion_func: Custom conversion service
        :param dtlpy.entities.item.Item item: item entity
        :return: annotation list, errors
        """
        item_id, annotations = self._extract_annotations_from_file(
            from_format=from_format,
            file_path=file_path,
            item=item
        )

        converted_annotations = self.convert(
            to_format=to_format,
            from_format=from_format,
            annotations=annotations,
            conversion_func=conversion_func,
            item=item
        )

        annotations_list, errors = self._sort_annotations(annotations=converted_annotations)

        if annotations_list:
            if save_locally:
                if item_id is not None:
                    item = self.dataset.items.get(item_id=item_id)
                filename = os.path.split(file_path)[-1]
                filename_no_ext = os.path.splitext(filename)[0]
                save_to = os.path.join(save_to, filename_no_ext)
                self.save_to_file(save_to=save_to, to_format=to_format, annotations=annotations_list, item=item)
            elif upload and to_format == AnnotationFormat.DATALOOP:
                item.annotations.upload(annotations=annotations_list)

        if pbar is not None:
            pbar.update()

        return annotations_list, errors

    @staticmethod
    def _sort_annotations(annotations):
        errors = list()
        success = list()

        for ann in annotations:
            if isinstance(ann, str) and 'Error converting annotation' in ann:
                errors.append(ann)
            else:
                success.append(ann)

        return success, errors

    def save_to_file(self, save_to, to_format, annotations, item=None):
        """
        Save annotations to a file.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str save_to: path to save the result to
        :param to_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param list annotations: annotation list to convert
        :param dtlpy.entities.item.Item item: item entity
        """
        # what file format
        if self.save_to_format is None:
            if to_format.lower() in [AnnotationFormat.DATALOOP, AnnotationFormat.COCO]:
                self.save_to_format = 'json'
            elif to_format.lower() in [AnnotationFormat.YOLO]:
                self.save_to_format = 'txt'
            else:
                self.save_to_format = 'xml'

        # save
        # JSON #
        if self.save_to_format == 'json':
            # save json
            save_to = save_to + '.json'
            with open(save_to, "w") as f:
                json.dump(annotations, f, indent=2)

        # TXT #
        elif self.save_to_format == 'txt':
            # save txt
            save_to = save_to + '.txt'
            with open(save_to, "w") as f:
                for ann in annotations:
                    if ann is not None:
                        f.write(' '.join([str(x) for x in ann]) + '\n')

        # XML #
        elif self.save_to_format == 'xml':
            item = Converter.__get_item_shape(item=item)
            depth = item.metadata['system'].get('channels', 3)
            output_annotation = {
                'path': item.filename,
                'filename': os.path.basename(item.filename),
                'folder': os.path.basename(os.path.dirname(item.filename)),
                'width': item.width,
                'height': item.height,
                'depth': depth,
                'database': 'Unknown',
                'segmented': 0,
                'objects': annotations
            }
            save_to = save_to + '.xml'
            environment = Environment(loader=PackageLoader('dtlpy', 'assets'),
                                      keep_trailing_newline=True)
            annotation_template = environment.get_template(self.xml_template_path)
            with open(save_to, 'w') as file:
                content = annotation_template.render(**output_annotation)
                file.write(content)
        else:
            raise exceptions.PlatformException('400', 'Unknown file format to save to')

    def _check_formats(self, from_format, to_format, conversion_func):
        if from_format.lower() not in self.known_formats and conversion_func is None:
            raise exceptions.PlatformException(
                "400",
                "Unknown annotation format: {}, possible formats: {}".format(
                    from_format, self.known_formats
                ),
            )
        if to_format.lower() not in self.known_formats and conversion_func is None:
            raise exceptions.PlatformException(
                "400",
                "Unknown annotation format: {}, possible formats: {}".format(
                    to_format, self.known_formats
                ),
            )

    def convert(self,
                annotations,
                from_format: str,
                to_format: str,
                conversion_func=None,
                item=None):
        """
        Convert annotation list or single annotation.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.item.Item item: item entity
        :param list or AnnotationCollection annotations: annotations list to convert
        :param str from_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param str to_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param Callable conversion_func: Custom conversion service
        :return: the annotations
        """
        # check known format
        self._check_formats(from_format=from_format, to_format=to_format, conversion_func=conversion_func)

        # check annotations param type
        if isinstance(annotations, entities.AnnotationCollection):
            annotations = annotations.annotations
        if not isinstance(annotations, list):
            annotations = [annotations]

        if AnnotationFormat.DATALOOP not in [to_format, from_format]:
            raise exceptions.PlatformException(
                "400", "Can only convert from or to dataloop format"
            )

        # call method
        if conversion_func is None:
            if from_format == AnnotationFormat.DATALOOP:
                method = self.converter_dict[to_format]["to"]
            else:
                method = self.converter_dict[from_format]["from"]
        else:
            method = self.custom_format

        # run all annotations
        for i_annotation, annotation in enumerate(annotations):
            self._convert_single(
                annotation=annotation,
                i_annotation=i_annotation,
                conversion_func=conversion_func,
                annotations=annotations,
                from_format=AnnotationFormat.DATALOOP,
                item=item,
                method=method
            )
        return annotations

    @staticmethod
    def _convert_single(method, **kwargs):
        annotations = kwargs.get('annotations', None)
        i_annotation = kwargs.get('i_annotation', None)
        try:
            ann = method(**kwargs)
            if annotations is not None and isinstance(i_annotation, int) and ann is not None:
                annotations[i_annotation] = ann
        except Exception:
            if annotations is not None and isinstance(i_annotation, int):
                annotations[i_annotation] = 'Error converting annotation: {}'.format(traceback.format_exc())
            else:
                raise

    @staticmethod
    def from_voc(annotation, **_):
        """
        Convert from VOC format to DATALOOP format. Use this as conversion_func for functions that ask for this param.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param annotation: annotations to convert
        :return: converted Annotation entity
        :rtype: dtlpy.entities.annotation.Annotation
        """
        bndbox = annotation.find('bndbox')

        if bndbox is None:
            raise Exception('No bndbox field found in annotation object')

        bottom = float(bndbox.find('ymax').text)
        top = float(bndbox.find('ymin').text)
        left = float(bndbox.find('xmin').text)
        right = float(bndbox.find('xmax').text)
        label = annotation.find('name').text

        if annotation.find('segmented'):
            if annotation.find('segmented') == '1':
                logger.warning('Only bounding box conversion is supported in voc format. Segmentation will be ignored.')

        attributes = list(annotation)
        attrs = dict()
        for attribute in attributes:
            if attribute.tag not in ['bndbox', 'name'] and len(list(attribute)) == 0:
                if attribute.text is not None:
                    if attribute.tag == 'attributes':
                        attribute_value = eval(attribute.text)
                        if isinstance(attribute_value, list):
                            attrs = {attribute_value[i]: i for i in range(len(attribute_value))}
                        else:
                            attrs.update(attribute_value)
                    else:
                        try:
                            attr_value = eval(attribute.text)
                        except:
                            attr_value = attribute.text
                        attrs[attribute.tag] = attr_value

        ann_def = entities.Box(label=label, top=top, bottom=bottom, left=left, right=right)
        new_annotation = entities.Annotation.new(annotation_definition=ann_def)
        if attrs is not None:
            try:
                new_annotation.attributes = attrs
            except:
                new_annotation.attributes = list(attrs.keys())
        return new_annotation

    def from_yolo(self, annotation, item=None, **kwargs):
        """
        Convert from YOLO format to DATALOOP format. Use this as conversion_func param for functions that ask for this param.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param annotation: annotations to convert
        :param dtlpy.entities.item.Item item: item entity
        :param kwargs: additional params
        :return: converted Annotation entity
        :rtype: dtlpy.entities.annotation.Annotation
        """
        (label_id, x, y, w, h) = annotation
        label_id = int(label_id)

        item = Converter.__get_item_shape(item=item)
        height = kwargs.get('height', None)
        width = kwargs.get('width', None)

        if height is None or width is None:
            if item is None or item.width is None or item.height is None:
                raise Exception('Need item width and height in order to convert yolo annotation to dataloop')
            height = item.height
            width = item.width

        x_center = x * width
        y_center = y * height
        w = w * width
        h = h * height

        top = y_center - (h / 2)
        bottom = y_center + (h / 2)
        left = x_center - (w / 2)
        right = x_center + (w / 2)

        label = self._label_by_category_id(category_id=label_id)

        ann_def = entities.Box(label=label, top=top, bottom=bottom, left=left, right=right)
        return entities.Annotation.new(annotation_definition=ann_def, item=item)

    def to_yolo(self, annotation, item=None, **_):
        """
        Convert from DATALOOP format to YOLO format. Use this as conversion_func param for functions that ask for this param.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.annotation.Annotation or dict annotation: annotations to convert
        :param dtlpy.entities.item.Item item: item entity
        :param **_: additional params
        :return: converted Annotation
        :rtype: tuple
        """
        if not isinstance(annotation, entities.Annotation):
            if item is None:
                item = self.dataset.items.get(item_id=annotation['itemId'])
            annotation = entities.Annotation.from_json(_json=annotation, item=item)
        elif item is None:
            item = annotation.item

        if annotation.type != "box":
            raise Exception('Only box annotations can be converted')

        item = Converter.__get_item_shape(item=item)

        if item is None or item.width is None or item.height is None:
            raise Exception("Cannot get item's width and height")

        width, height = (item.width, item.height)
        if item.system.get('exif', {}).get('Orientation', 0) in [5, 6, 7, 8]:
            width, height = (item.height, item.width)

        dw = 1.0 / width
        dh = 1.0 / height
        x = (annotation.left + annotation.right) / 2.0
        y = (annotation.top + annotation.bottom) / 2.0
        w = annotation.right - annotation.left
        h = annotation.bottom - annotation.top
        x = x * dw
        w = w * dw
        y = y * dh
        h = h * dh

        label_id = self.labels[annotation.label]

        ann = (label_id, x, y, w, h)
        return ann

    def _label_by_category_id(self, category_id):
        if len(self.labels) > 0:
            for label_name, label_index in self.labels.items():
                if label_index == category_id:
                    return label_name
        raise Exception('label category id not found: {}'.format(category_id))

    def from_coco(self, annotation, **kwargs):
        """
        Convert from COCO format to DATALOOP format. Use this as conversion_func param for functions that ask for this param.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param annotation: annotations to convert
        :param kwargs: additional params
        :return: converted Annotation entity
        :rtype: dtlpy.entities.annotation.Annotation
        """
        item = kwargs.get('item', None)
        _id = annotation.get('id', None)
        category_id = annotation.get('category_id', None)
        segmentation = annotation.get('segmentation', None)
        iscrowd = annotation.get('iscrowd', None)
        label = self._label_by_category_id(category_id=category_id)
        ann_def = None

        if hasattr(self, '_only_bbox') and self._only_bbox:
            bbox = annotation.get('bbox', None)
            left = bbox[0]
            top = bbox[1]
            right = left + bbox[2]
            bottom = top + bbox[3]

            ann_def = entities.Box(top=top, left=left, bottom=bottom, right=right, label=label)
        else:
            if iscrowd is not None and int(iscrowd) == 1:
                ann_def = entities.Segmentation(label=label, geo=COCOUtils.rle_to_binary_mask(rle=segmentation))
            else:
                if segmentation is not None and len(segmentation) == 1:
                    segmentation = segmentation[0]
                    if segmentation:
                        ann_def = entities.Polygon(label=label,
                                                   geo=COCOUtils.rle_to_binary_polygon(segmentation=segmentation))

                elif segmentation is not None and len(segmentation) > 1:
                    # TODO - support conversion of split annotations
                    raise exceptions.PlatformException('400',
                                                       'unable to convert.'
                                                       'Converter does not support split annotations: {}'.format(
                                                           _id))
                if not segmentation:
                    bbox = annotation.get('bbox', None)
                    if bbox:
                        left = bbox[0]
                        top = bbox[1]
                        right = left + bbox[2]
                        bottom = top + bbox[3]
                        ann_def = entities.Box(top=top, left=left, bottom=bottom, right=right, label=label)
                    else:
                        raise Exception('Unable to convert annotation, not coordinates')

        return entities.Annotation.new(annotation_definition=ann_def, item=item)

    @staticmethod
    def to_coco(annotation, item=None, **_):
        """
        Convert from DATALOOP format to COCO format. Use this as conversion_func param for functions that ask for this param.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.annotation.Annotation or dict annotation: annotations to convert
        :param dtlpy.entities.item.Item item: item entity
        :param **_: additional params
        :return: converted Annotation
        :rtype: dict
        """
        item = Converter.__get_item_shape(item=item)
        height = item.height if item is not None else None
        width = item.width if item is not None else None

        if not isinstance(annotation, entities.Annotation):
            annotation = entities.Annotation.from_json(annotation, item=item)
        area = 0
        iscrowd = 0
        segmentation = [[]]
        if annotation.type == 'polyline':
            raise Exception('Unable to convert annotation of type "polyline" to coco')

        if annotation.type in ['binary', 'segment']:
            if height is None or width is None:
                raise Exception(
                    'Item must have height and width to convert {!r} annotation to coco'.format(annotation.type))

        # build annotation
        keypoints = None
        if annotation.type in ['binary', 'box', 'segment', 'pose']:
            x = annotation.left
            y = annotation.top
            w = annotation.right - x
            h = annotation.bottom - y
            area = float(h * w)
            if annotation.type == 'binary':
                # segmentation = COCOUtils.binary_mask_to_rle(binary_mask=annotation.geo, height=height, width=width)
                segmentation = COCOUtils.binary_mask_to_rle_encode(binary_mask=annotation.geo)
                area = int(annotation.geo.sum())
                iscrowd = 1
            elif annotation.type in ['segment']:
                segmentation, area = COCOUtils.polygon_to_rle(geo=annotation.geo, height=height, width=width)
            elif annotation.type == 'pose':
                keypoints = list()
                for point in annotation.annotation_definition.points:
                    keypoints.append(point.x)
                    keypoints.append(point.y)
                    if isinstance(point.attributes, list):
                        if 'visible' in point.attributes and \
                                ("not-visible" in point.attributes or 'not_visible' in point.attributes):
                            keypoints.append(0)
                        elif 'not-visible' in point.attributes or 'not_visible' in point.attributes:
                            keypoints.append(1)
                        elif 'visible' in point.attributes:
                            keypoints.append(2)
                        else:
                            keypoints.append(0)
                    else:
                        list_attributes = list(point.attributes.values())
                        if 'Visible' in list_attributes:
                            keypoints.append(2)
                        else:
                            keypoints.append(0)
                x_points = keypoints[0::3]
                y_points = keypoints[1::3]
                x0, x1, y0, y1 = np.min(x_points), np.max(x_points), np.min(y_points), np.max(y_points)
                area = float((x1 - x0) * (y1 - y0))
        else:
            raise Exception('Unable to convert annotation of type {} to coco'.format(annotation.type))

        ann = dict()
        ann['bbox'] = [float(x), float(y), float(w), float(h)]
        ann["segmentation"] = segmentation
        ann["area"] = area
        ann["iscrowd"] = iscrowd
        if keypoints is not None:
            ann["keypoints"] = keypoints
        return ann

    @staticmethod
    def to_voc(annotation, item=None, **_):
        """
        Convert from DATALOOP format to VOC format. Use this as conversion_func param for functions that ask for this param.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.annotation.Annotation or dict annotation: annotations to convert
        :param dtlpy.entities.item.Item item: item entity
        :param **_: additional params
        :return: converted Annotation
        :rtype: dict
        """
        if not isinstance(annotation, entities.Annotation):
            annotation = entities.Annotation.from_json(annotation, item=item)

        if annotation.type != 'box':
            raise exceptions.PlatformException('400', 'Annotation must be of type box')

        label = annotation.label

        attributes = dict()
        if annotation.attributes is not None:
            if isinstance(annotation.attributes, dict) and len(annotation.attributes) > 0:
                attributes = annotation.attributes
            elif isinstance(annotation.attributes, list) and len(annotation.attributes) > 0:
                attributes = {annotation.attributes[i]: i for i in range(len(annotation.attributes))}

        left = annotation.left
        top = annotation.top
        right = annotation.right
        bottom = annotation.bottom

        ann = {'name': label,
               'xmin': left,
               'ymin': top,
               'xmax': right,
               'ymax': bottom
               }
        if attributes:
            ann['attributes'] = attributes
        return ann

    @staticmethod
    def custom_format(annotation,
                      conversion_func,
                      i_annotation=None,
                      annotations=None,
                      from_format=None,
                      item=None, **_):
        """
        Custom convert function.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.annotation.Annotation or dict annotation: annotations to convert
        :param Callable conversion_func: Custom conversion service
        :param int i_annotation: annotation index
        :param list annotations: list of annotations
        param str from_format: AnnotationFormat to convert to – AnnotationFormat.COCO, AnnotationFormat.YOLO, AnnotationFormat.VOC, AnnotationFormat.DATALOOP
        :param dtlpy.entities.item.Item item: item entity
        :return: converted Annotation
        """
        if from_format == AnnotationFormat.DATALOOP and isinstance(annotation, dict):
            annotation = entities.Annotation.from_json(_json=annotation, item=item)

        ann = conversion_func(annotation, item)

        if isinstance(annotations[i_annotation], entities.Annotation) and annotations[i_annotation].item is None:
            ann.item = item

        return ann



from . import BaseAnnotationDefinition


class GisType:
    """
    State enum
    """
    BOX = 'box'
    POLYGON = 'polygon'
    POLYLINE = 'polyline'
    POINT = 'point'


class Gis(BaseAnnotationDefinition):
    """
        Box annotation object
        Can create a box using 2 point using: "top", "left", "bottom", "right" (to form a box [(left, top), (right, bottom)])
        For rotated box add the "angel"
    """
    type = "gis"

    def __init__(self,
                 annotation_type: GisType,
                 geo,
                 label=None,
                 attributes=None,
                 description=None,
                 ):
        """
        Can create gis annotation using points:

        :param geo: list of points
        :param label: annotation label
        :param attributes: a list of attributes for the annotation
        :param description:

        :return:
        """
        super().__init__(description=description, attributes=attributes)

        if geo is None:
            raise ValueError('geo must be provided')
        if annotation_type is None:
            raise ValueError('annotation_type must be provided')
        self.label = label
        self.annotation = None
        self.geo = geo
        self.annotation_type = annotation_type

    def to_coordinates(self, color):
        return {
            "geo_type": self.annotation_type,
            "wgs84_geo_coordinates": self.geo
        }

    @classmethod
    def from_json(cls, _json):
        json_coordinates = _json.get("coordinates", {}) if "coordinates" in _json else _json.get("data", {})
        coordinates = json_coordinates.get("wgs84_geo_coordinates", None)
        annotations_type = json_coordinates.get("geo_type", None)
        if coordinates is None:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        return cls(
            annotation_type=annotations_type,
            geo=coordinates,
            label=_json["label"],
            attributes=_json.get("attributes", None)
        )



import logging
import time
import numpy as np

from .. import entities, repositories, exceptions, miscellaneous, services, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')
MAX_SLEEP_TIME = 30


class PipelineExecutions:
    """
    PipelineExecutions Repository

    The PipelineExecutions class allows users to manage pipeline executions. See our documentation for more information on `pipelines <https://dataloop.ai/docs/pipelines-overview>`_.
    """

    def __init__(
            self,
            client_api: ApiClient,
            project: entities.Project = None,
            pipeline: entities.Pipeline = None
    ):
        self._client_api = client_api
        self._project = project
        self._pipeline = pipeline

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            try:
                self._project = repositories.Projects(client_api=self._client_api).get()
            except exceptions.NotFound:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Missing "project". need to set a Project entity or use project.pipelines repository')
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def pipeline(self) -> entities.Pipeline:
        assert isinstance(self._pipeline, entities.Pipeline)
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline: entities.Pipeline):
        if not isinstance(pipeline, entities.Pipeline):
            raise ValueError('Must input a valid pipeline entity')
        self._pipeline = pipeline

    ###########
    # methods #
    ###########
    # @_api_reference.add(path='/pipelines/{pipelineId}/executions/{executionId}', method='get')
    def get(self,
            pipeline_execution_id: str,
            pipeline_id: str = None
            ) -> entities.PipelineExecution:
        """
        Get Pipeline Execution object

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str pipeline_execution_id: pipeline execution id
        :param str pipeline_id: pipeline id
        :return: PipelineExecution object
        :rtype: dtlpy.entities.pipeline_execution.PipelineExecution

        **Example**:

        .. code-block:: python

            pipeline_executions = pipeline.pipeline_executions.get(pipeline_id='pipeline_id')
        """

        if pipeline_id is None and self._pipeline is None:
            raise exceptions.PlatformException('400', 'Must provide param pipeline_id')
        elif pipeline_id is None:
            pipeline_id = self._pipeline.id

        success, response = self._client_api.gen_request(
            req_type="get",
            path="/pipelines/{pipeline_id}/executions/{pipeline_execution_id}".format(
                pipeline_id=pipeline_id,
                pipeline_execution_id=pipeline_execution_id
            )
        )
        if not success:
            raise exceptions.PlatformException(response)

        pipeline_execution = entities.PipelineExecution.from_json(
            client_api=self._client_api,
            _json=response.json(),
            pipeline=self._pipeline
        )

        return pipeline_execution

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.PipelineExecution]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]

        for i_pipeline_execution, pipeline_execution in enumerate(response_items):
            jobs[i_pipeline_execution] = pool.submit(
                entities.PipelineExecution._protected_from_json,
                **{
                    'client_api': self._client_api,
                    '_json': pipeline_execution,
                    'pipeline': self._pipeline
                }
            )

        # get all results
        # noinspection PyUnresolvedReferences
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        pipeline_executions = miscellaneous.List([r[1] for r in results if r[0] is True])
        return pipeline_executions

    def _list(self, filters: entities.Filters):
        url = '/pipelines/query'

        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path=url,
            json_req=filters.prepare()
        )
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/pipelines/query', method='post')
    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List project pipeline executions.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            pipeline_executions = pipeline.pipeline_executions.list()
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.PIPELINE_EXECUTION)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.PIPELINE_EXECUTION:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.PIPELINE_EXECUTION. Got: {!r}'.format(
                    filters.resource))

        project_id = None
        if self._project is not None:
            project_id = self._project.id

        # TODO - uncomment this after DAT-24496 is done and cycles have projectId
        # if self._project is not None:
        #     filters.add(field='projectId', values=self.project.id)

        if self._pipeline is not None:
            filters.add(field='pipelineId', values=self._pipeline.id)

        paged = entities.PagedEntities(
            items_repository=self,
            filters=filters,
            page_offset=filters.page,
            page_size=filters.page_size,
            project_id=project_id,
            client_api=self._client_api
        )

        paged.get_page()
        return paged

    @_api_reference.add(path="/pipelines/{pipelineId}/execute", method="post")
    def create(
            self,
            pipeline_id: str = None,
            execution_input=None,
            node_id: str = None,
            test_mode: bool = None
    ):
        """
        Execute a pipeline.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param pipeline_id: pipeline id
        :param execution_input: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}
        :param node_id: node id to start from
        :param bool test_mode: if True, the pipeline will be executed in test mode
        :return: entities.PipelineExecution object
        :rtype: dtlpy.entities.pipeline_execution.PipelineExecution

        **Example**:

        .. code-block:: python

            pipeline_execution = pipeline.pipeline_executions.create(pipeline_id='pipeline_id', execution_input={'item': 'item_id'})
        """
        if pipeline_id is None:
            if self._pipeline is None:
                raise exceptions.PlatformException('400', 'Please provide pipeline id')
            pipeline_id = self._pipeline.id

        pipeline_options = dict()
        if test_mode is not None:
            pipeline_options["testMode"] = test_mode
        if execution_input is None:
            # support pipeline executions without any input
            pass
        elif isinstance(execution_input, dict):
            pipeline_options["input"] = execution_input
        else:
            if not isinstance(execution_input, list):
                execution_input = [execution_input]
            if len(execution_input) > 0 and isinstance(execution_input[0], entities.FunctionIO):
                pipeline_options["input"] = dict()
                for single_input in execution_input:
                    pipeline_options["input"].update(single_input.to_json(resource="execution"))
            else:
                raise exceptions.PlatformException('400', 'Unknown input type')

        if node_id is not None:
            pipeline_options["nodeId"] = node_id

        success, response = self._client_api.gen_request(
            path="/pipelines/{}/execute".format(pipeline_id),
            req_type="POST",
            json_req={"pipeline": pipeline_options},
        )

        if not success:
            raise exceptions.PlatformException(response)

        execution = entities.PipelineExecution.from_json(_json=response.json(),
                                                         client_api=self._client_api,
                                                         pipeline=self._pipeline)
        return execution

    @_api_reference.add(path='/pipelines/{pipelineId}/execute', method='post')
    def create_batch(
            self,
            pipeline_id: str,
            filters,
            execution_inputs=None,
            wait=True,
            node_id: str = None
    ):
        """
        Create batch executions of a pipeline.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param pipeline_id: pipeline id
        :param filters: Filters entity for a filtering before execute
        :param execution_inputs: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}, that represent the extra inputs of the function
        :param bool wait: wait until create task finish
        :return: entities.PipelineExecution object
        :rtype: dtlpy.entities.pipeline_execution.PipelineExecution

        **Example**:

        .. code-block:: python

            command = pipeline.pipeline_executions.create_batch(
                        execution_inputs=dl.FunctionIO(type=dl.PackageInputType.STRING, value='test', name='string'),
                        filters=dl.Filters(field='dir', values='/test'))
        """
        if pipeline_id is None:
            if self._pipeline is None:
                raise exceptions.PlatformException('400', 'Please provide pipeline id')
            pipeline_id = self._pipeline.id

        if filters is None:
            raise exceptions.PlatformException('400', 'Please provide filter')
        extra_input = dict()

        if execution_inputs is None:
            execution_inputs = {}

        if isinstance(execution_inputs, dict):
            extra_input = execution_inputs
        else:
            if not isinstance(execution_inputs, list):
                execution_inputs = [execution_inputs]
            if len(execution_inputs) > 0 and isinstance(execution_inputs[0], entities.FunctionIO):
                for single_input in execution_inputs:
                    extra_input.update(single_input.to_json(resource='execution'))
            else:
                raise exceptions.PlatformException('400', 'Unknown input type')
        payload = dict()
        payload['batch'] = dict()
        payload['batch']['query'] = filters.prepare()
        payload['batch']['args'] = extra_input

        if node_id is not None:
            payload['nodeId'] = node_id

        success, response = self._client_api.gen_request(
            path='/pipelines/{}/execute'.format(pipeline_id),
            req_type='POST',
            json_req=payload
        )

        if not success:
            raise exceptions.PlatformException(response)

        response_json = response.json()
        command = entities.Command.from_json(_json=response_json,
                                             client_api=self._client_api)
        if wait:
            command = command.wait(timeout=0)
        return command

    @_api_reference.add(path='/pipelines/{pipelineId}/executions/rerun', method='post')
    def rerun(self,
              pipeline_id: str = None,
              method: str = None,
              start_nodes_ids: list = None,
              filters: entities.Filters = None,
              wait: bool = True
              ) -> bool:
        """
        Get Pipeline Execution object

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str pipeline_id: pipeline id
        :param str method: method to run
        :param list start_nodes_ids: list of start nodes ids
        :param filters: Filters entity for a filtering before execute
        :param bool wait: wait until rerun finish
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            pipeline.pipeline_executions.rerun(pipeline_id='pipeline_id', method=dl.CycleRerunMethod.START_FROM_BEGINNING)
        """

        if pipeline_id is None and self._pipeline is None:
            raise exceptions.PlatformException('400', 'Must provide param pipeline_id')
        elif pipeline_id is None:
            pipeline_id = self._pipeline.id

        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.PIPELINE_EXECUTION)

        success, response = self._client_api.gen_request(
            req_type="post",
            path="/pipelines/{pipeline_id}/executions/rerun".format(
                pipeline_id=pipeline_id,
            ),
            json_req={
                'pipeline': {
                    "method": method,
                    "startNodeIds": start_nodes_ids,
                },
                "batch": {
                    "query": filters.prepare()
                }
            }
        )

        if not success:
            raise exceptions.PlatformException(response)

        command = entities.Command.from_json(_json=response.json(),
                                             client_api=self._client_api)
        if not wait:
            return command
        command = command.wait(timeout=0)

        if 'cycleOptions' not in command.spec:
            raise exceptions.PlatformException(error='400',
                                               message="cycleOptions key is missing in command response: {!r}"
                                               .format(response))
        return True

    def wait(self,
             pipeline_execution_id: str = None,
             pipeline_execution: entities.PipelineExecution = None,
             timeout: int = None,
             backoff_factor=1):
        """
        Get Service execution object.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str pipeline_execution_id: pipeline execution id
        :param str pipeline_execution: dl.PipelineExecution, optional. must input one of pipeline execution or pipeline_execution_id
        :param int timeout: seconds to wait until TimeoutError is raised. if <=0 - wait until done - by default wait take the service timeout
        :param float backoff_factor: A backoff factor to apply between attempts after the second try
        :return: Service execution object
        :rtype: dtlpy.entities.pipeline_execution.PipelineExecution

        **Example**:

        .. code-block:: python

            pipeline.pipeline_executions.wait(pipeline_execution_id='pipeline_execution_id')
        """
        if pipeline_execution is None:
            if pipeline_execution_id is None:
                raise ValueError('Must input at least one: [pipeline_execution, pipeline_execution_id]')
        else:
            pipeline_execution_id = pipeline_execution.id
        elapsed = 0
        start = time.time()
        if timeout is None or timeout <= 0:
            timeout = np.inf

        num_tries = 1
        while elapsed < timeout:
            pipeline_execution = self.get(pipeline_execution_id=pipeline_execution_id)
            if not pipeline_execution.in_progress():
                break
            elapsed = time.time() - start
            if elapsed >= timeout:
                raise TimeoutError(
                    f"Pipeline execution wait() function timed out. id: {pipeline_execution.id!r}, status: {pipeline_execution.status}.")
            sleep_time = np.min([timeout - elapsed, backoff_factor * (2 ** num_tries), MAX_SLEEP_TIME])
            num_tries += 1
            logger.debug(
                f"Pipeline execution {pipeline_execution.id!r} has been running for {elapsed:.2f}[s]. Sleeping for {sleep_time:.2f}[s]")
            time.sleep(sleep_time)
        return pipeline_execution



def main():
    import dtlpy as dl

    ########
    # prep #
    ########
    project = dl.projects.get(project_name='RQL')
    dataset = project.datasets.get(dataset_name='Dataset')

    #################
    ###   Items   ###
    #################

    ##################
    # create filters #
    ##################
    filters = dl.Filters(resource=dl.FiltersResource.ITEM)  # set resource - optional - default is item
    # add filter - only files
    filters.add(field='type', values='file')
    # add filter - only annotated items
    filters.add(field='annotated', values=True)
    # add filter - filename includes 'dog'
    filters.add(field='filename', values='/dog.jpg')
    # -- time filters -- must be in ISO format and in UTC (offset from local time). converting using datetime package as follows:
    import datetime, time
    timestamp = datetime.datetime(year=2019, month=10, day=27, hour=15, minute=39, second=6,
                                  tzinfo=datetime.timezone(datetime.timedelta(seconds=-time.timezone))).isoformat()
    filters.add(field='createdAt', values=timestamp, operator=dl.FiltersOperations.GREATER_THAN)

    ######################
    # get filtered items #
    ######################
    # return results sorted by ascending id 
    filters.sort_by(field='filename')
    pages = dataset.items.list(filters=filters)

    #########################
    # update filtered items #
    #########################
    # to add filed annotatedDogs to all filtered items and give value True
    # this field will be added to user metadata
    # create update order
    update_values = {'annotatedDogsSingJune2019': True}

    # update
    pages = dataset.items.update(filters=filters, update_values=update_values)

    # #########################
    # # delete filtered items #
    # #########################
    # dataset.items.delete(filters=filters)

    #####################################
    # filter items by their annotations #
    #####################################
    filters = dl.Filters()
    # set resource
    filters.resource = 'items'
    # add filter - only files
    filters.add(field='type', values='file')

    # add annotation filters - only items with 'box' annotations
    filters.add_join(field='type', values='box')

    # get results
    pages = dataset.items.list(filters=filters)

    #######################
    ###   Annotations   ###
    #######################

    ##################
    # create filters #
    ##################
    filters = dl.Filters()
    # set resource
    filters.resource = dl.FiltersResource.ANNOTATION
    # add filter - only box annotations
    filters.add(field='type', values='box')
    # add filter - only dogs
    filters.add(field='label', values=['Dog', 'cat'], operator=dl.FiltersOperations.IN)
    # add filter - annotated by Joe and David
    filters.add(field='creator',
                values=['Joe@dataloop.ai', 'David@dataloop.ai', 'oa-test-1@dataloop.ai'],
                operator=dl.FiltersOperations.IN
                )

    ############################
    # get filtered annotations #
    ############################
    # return results sorted by descending id 
    filters.sort_by(field='id', value=dl.FiltersOrderByDirection.DESCENDING)
    pages = dataset.items.list(filters=filters)

    ###############################
    # update filtered annotations #
    ###############################
    # to add filed annotation_quality to all filtered annotations and give them value 'high'
    # this field will be added to user metadata
    # create update order
    update_values = {'annotation_quality': 'high'}

    # update
    pages = dataset.items.update(filters=filters, update_values=update_values)

    ###############################
    # delete filtered annotations #
    ###############################
    dataset.items.delete(filters=filters)


if __name__ == '__main__':
    main()
